# PyWCSTools - a simple SWIG wrapping of WCSTools (http://tdc-www.harvard.edu/software/wcstools/) by Doug Mink
#
# (c) 2007, 2008 Matt Hilton
#
# See the README file for information on usage and distribution.


"""Module for producing astronomical plots.

(c) 2007-2024 Matt Hilton

This module provides the matplotlib powered ImagePlot class, which is designed to be flexible. 
ImagePlots can have RA, Dec. coordinate axes, contour overlays, and have objects marked in them, 
using WCS coordinates. RGB plots are supported too.

"""

import math
from . import astImages
from . import astWCS
from . import astCoords
import numpy as np
import astropy.io.fits as pyfits
from scipy import interpolate
import pylab
import matplotlib.patches as patches
import sys

# Handle unicode python 2 and 3
if sys.version < '3':
    import codecs
    def u(x):
        return codecs.unicode_escape_decode(x)[0]
else:
    def u(x):
        return x
    
DEC_TICK_STEPS=[{'deg': 1.0/60.0/60.0,  'unit': "s"}, 
                {'deg': 2.0/60.0/60.0,  'unit': "s"},
                {'deg': 5.0/60.0/60.0,  'unit': "s"}, 
                {'deg': 10.0/60.0/60.0, 'unit': "s"},
                {'deg': 30.0/60.0/60.0, 'unit': "s"},
                {'deg': 1.0/60.0,       'unit': "m"},
                {'deg': 2.0/60.0,       'unit': "m"},
                {'deg': 5.0/60.0,       'unit': "m"},
                {'deg': 15.0/60.0,      'unit': "m"},
                {'deg': 30.0/60.0,      'unit': "m"}, 
                {'deg': 1.0,            'unit': "d"},
                {'deg': 2.0,            'unit': "d"},
                {'deg': 4.0,            'unit': "d"},
                {'deg': 5.0,            'unit': "d"},
                {'deg': 10.0,           'unit': "d"},
                {'deg': 20.0,           'unit': "d"},
                {'deg': 30.0,           'unit': "d"}]
"""Defines the possible coordinate label steps on the delination axis in
sexagesimal mode. Dictionary format: {'deg', 'unit'}"""

RA_TICK_STEPS=[ {'deg': (0.5/60.0/60.0/24.0)*360.0,  'unit': "s"},
                {'deg': (1.0/60.0/60.0/24.0)*360.0,  'unit': "s"},
                {'deg': (2.0/60.0/60.0/24.0)*360.0,  'unit': "s"}, 
                {'deg': (4.0/60.0/60.0/24.0)*360.0,  'unit': "s"}, 
                {'deg': (5.0/60.0/60.0/24.0)*360.0,  'unit': "s"}, 
                {'deg': (10.0/60.0/60.0/24.0)*360.0, 'unit': "s"},
                {'deg': (20.0/60.0/60.0/24.0)*360.0, 'unit': "s"},
                {'deg': (30.0/60.0/60.0/24.0)*360.0, 'unit': "s"},
                {'deg': (1.0/60.0/24.0)*360.0,       'unit': "m"},
                {'deg': (2.0/60.0/24.0)*360.0,       'unit': "m"},
                {'deg': (5.0/60.0/24.0)*360.0,       'unit': "m"},
                {'deg': (10.0/60.0/24.0)*360.0,      'unit': "m"},
                {'deg': (20.0/60.0/24.0)*360.0,      'unit': "m"},
                {'deg': (30.0/60.0/24.0)*360.0,      'unit': "m"}, 
                {'deg': (1.0/24.0)*360.0,            'unit': "h"},
                {'deg': (3.0/24.0)*360.0,            'unit': "h"},
                {'deg': (6.0/24.0)*360.0,            'unit': "h"},
                {'deg': (12.0/24.0)*360.0,           'unit': "h"}]
"""Defines the possible coordinate label steps on the right ascension axis in
sexagesimal mode. Dictionary format: {'deg', 'unit'}"""

DECIMAL_TICK_STEPS=[0.001, 0.0025, 0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.0, 2.5, 5.0, 10.0, 30.0, 90.0]
"""Defines the possible coordinate label steps on both coordinate axes in decimal degrees mode."""

DEG = u("\N{DEGREE SIGN}")

PRIME = "$^\prime$"

DOUBLE_PRIME = "$^{\prime\prime}$"

#---------------------------------------------------------------------------------------------------
class ImagePlot:
    """This class describes a matplotlib image plot containing an astronomical image with an
    associated WCS.
    
    Objects within the image boundaries can be marked by passing their WCS coordinates to 
    L{ImagePlot.addPlotObjects}.
    
    Other images can be overlaid using L{ImagePlot.addContourOverlay}.
    
    For images rotated with North at the top, East at the left (as can be done using
    L{astImages.clipRotatedImageSectionWCS} or L{astImages.resampleToTanProjection}, WCS coordinate
    axes can be plotted, with tick marks set appropriately for the image size. Otherwise, a compass 
    can be plotted showing the directions of North and East in the image.

    RGB images are also supported.
    
    The plot can of course be tweaked further after creation using matplotlib/pylab commands.
    
    """
    def __init__(self, imageData, imageWCS, axes = [0.1,0.1,0.8,0.8], \
        cutLevels = ["smart", 99.5], colorMapName = "gray", title = None, axesLabels = "sexagesimal", \
        axesFontFamily="serif", axesFontSize=12.0, RATickSteps="auto", decTickSteps="auto",
        colorBar = False, interpolation = "bilinear"):
        """Makes an ImagePlot from the given image array and astWCS. For coordinate axes to work, the
        image and WCS should have been rotated such that East is at the left, North is at the top
        (see e.g. L{astImages.clipRotatedImageSectionWCS}, or L{astImages.resampleToTanProjection}).
        
        If imageData is given as a list in the format [r, g, b], a color RGB plot will be made. However,
        in this case the cutLevels must be specified manually for each component as a list -
        i.e. cutLevels = [[r min, r max], [g min, g max], [b min, b max]]. In this case of course, the
        colorMap will be ignored. All r, g, b image arrays must have the same dimensions.
        
        Set axesLabels = None to make a plot without coordinate axes plotted.
        
        The axes can be marked in either sexagesimal or decimal celestial coordinates. If RATickSteps
        or decTickSteps are set to "auto", the appropriate axis scales will be determined automatically 
        from the size of the image array and associated WCS. The tick step sizes can be overidden. 
        If the coordinate axes are in sexagesimal format a dictionary in the format {'deg', 'unit'} is 
        needed (see L{RA_TICK_STEPS} and L{DEC_TICK_STEPS} for examples). If the coordinate axes are in
        decimal format, the tick step size is specified simply in RA, dec decimal degrees.
        
        @type imageData: np array or list
        @param imageData: image data array or list of np arrays [r, g, b]
        @type imageWCS: astWCS.WCS
        @param imageWCS: astWCS.WCS object
        @type axes: list
        @param axes: specifies where in the current figure to draw the finder chart (see pylab.axes)
        @type cutLevels: list
        @param cutLevels: sets the image scaling - available options:
            - pixel values: cutLevels=[low value, high value].
            - histogram equalisation: cutLevels=["histEq", number of bins ( e.g. 1024)]
            - relative: cutLevels=["relative", cut per cent level (e.g. 99.5)]
            - smart: cutLevels=["smart", cut per cent level (e.g. 99.5)]
        ["smart", 99.5] seems to provide good scaling over a range of different images.
        Note that for RGB images, cut levels must be specified manually i.e. as a list:
        [[r min, rmax], [g min, g max], [b min, b max]]
        @type colorMapName: string
        @param colorMapName: name of a standard matplotlib colormap, e.g. "hot", "cool", "gray"
        etc. (do "help(pylab.colormaps)" in the Python interpreter to see available options)
        @type title: string
        @param title: optional title for the plot
        @type axesLabels: string
        @param axesLabels: either "sexagesimal" (for H:M:S, D:M:S), "decimal" (for decimal degrees)
        or None (for no coordinate axes labels)
        @type axesFontFamily: string
        @param axesFontFamily: matplotlib fontfamily, e.g. 'serif', 'sans-serif' etc.
        @type axesFontSize: float
        @param axesFontSize: font size of axes labels and titles (in points)
        @type colorBar: bool
        @param colorBar: if True, plot a vertical color bar at the side of the image indicating the intensity
        scale.
        @type interpolation: string
        @param interpolation: interpolation to apply to the image plot (see the documentation for
                              the matplotlib.pylab.imshow command)
        
        """
                
        self.RADeg, self.decDeg=imageWCS.getCentreWCSCoords()
        self.wcs=imageWCS
        
        # Handle case where imageData is [r, g, b]
        if type(imageData) == list:
            if len(imageData) == 3:
                if len(cutLevels) == 3:
                    r=astImages.normalise(imageData[0], cutLevels[0])
                    g=astImages.normalise(imageData[1], cutLevels[1])
                    b=astImages.normalise(imageData[2], cutLevels[2])
                    rgb=np.array([r.transpose(), g.transpose(), b.transpose()])
                    rgb=rgb.transpose()
                    self.data=rgb
                    self.rgbImage=True
                else:
                    raise Exception("tried to create a RGB array, but cutLevels is not a list of 3 lists")

            else:
                raise Exception("tried to create a RGB array but imageData is not a list of 3 arrays")
        else:
            self.data=imageData
            self.rgbImage=False
        
        self.axes=pylab.axes(axes)
        self.cutLevels=cutLevels
        self.colorMapName=colorMapName
        self.title=title
        self.axesLabels=axesLabels
        self.colorBar=colorBar
        self.axesFontSize=axesFontSize
        self.axesFontFamily=axesFontFamily
        
        self.flipXAxis=False
        self.flipYAxis=False
                
        self.interpolation=interpolation
        
        if self.axesLabels != None:
            
            # Allow user to override the automatic coord tick spacing
            if self.axesLabels == "sexagesimal":
                if RATickSteps != "auto":
                    if type(RATickSteps) != dict or "deg" not in list(RATickSteps.keys()) \
                            or "unit" not in list(RATickSteps.keys()):
                        raise Exception("RATickSteps needs to be in format {'deg', 'unit'} for sexagesimal axes labels")
                if decTickSteps != "auto":
                    if type(decTickSteps) != dict or "deg" not in list(decTickSteps.keys()) \
                            or "unit" not in list(decTickSteps.keys()):
                        raise Exception("decTickSteps needs to be in format {'deg', 'unit'} for sexagesimal axes labels")
            elif self.axesLabels == "decimal":
                if RATickSteps != "auto":
                    if type(RATickSteps) != float:
                        raise Exception("RATickSteps needs to be a float (if not 'auto') for decimal axes labels")
                if decTickSteps != "auto":
                    if type(decTickSteps) != float:
                        raise Exception("decTickSteps needs to be a float (if not 'auto') for decimal axes labels")
            self.RATickSteps=RATickSteps
            self.decTickSteps=decTickSteps
        
            self.calcWCSAxisLabels(axesLabels = self.axesLabels)
        
        # this list stores objects to overplot, add to it using addPlotObjects()
        self.plotObjects=[] 
        
        # this list stores image data to overlay as contours, add to it using addContourOverlay()
        self.contourOverlays=[]
        
        self.draw()


    def draw(self):
        """Redraws the ImagePlot.
        
        """
        
        pylab.axes(self.axes)
        pylab.cla()
        
        if self.title != None:
            pylab.title(self.title)
        try:
            colorMap=pylab.cm.get_cmap(self.colorMapName)
        except AssertionError:
            raise Exception(self.colorMapName+"is not a defined matplotlib colormap.")
        
        if self.rgbImage == False:
            self.cutImage=astImages.intensityCutImage(self.data, self.cutLevels)
            if self.cutLevels[0]=="histEq":
                pylab.imshow(self.cutImage['image'],  interpolation=self.interpolation, origin='lower', cmap=colorMap)
            else:
                pylab.imshow(self.cutImage['image'],  interpolation=self.interpolation,  norm=self.cutImage['norm'], \
                            origin='lower', cmap=colorMap)
        else:
            pylab.imshow(self.data, interpolation="bilinear", origin='lower')
        
        if self.colorBar == True:
            pylab.colorbar(shrink=0.8)
        
        for c in self.contourOverlays:
            pylab.contour(c['contourData']['scaledImage'], c['contourData']['contourLevels'], 
                            colors=c['color'], linewidths=c['width'])
        
        for p in self.plotObjects:
            for x, y, l in zip(p['x'], p['y'], p['objLabels']):
                if p['symbol'] == "circle":
                    c=patches.Circle((x, y), radius=p['sizePix']/2.0, fill=p['fill'], color=p['color'],
                                        linewidth=p['width'])
                    self.axes.add_patch(c)
                elif p['symbol'] == "box":
                    c=patches.Rectangle((x-p['sizePix']/2, y-p['sizePix']/2), p['sizePix'], p['sizePix'], 
                        fill=p['fill'], color=p['color'], linewidth=p['width'])
                    self.axes.add_patch(c)
                elif p['symbol'] == "cross":
                    pylab.plot([x-p['sizePix']/2, x+p['sizePix']/2], [y, y], linestyle='-', 
                        linewidth=p['width'], color= p['color'])
                    pylab.plot([x, x], [y-p['sizePix']/2, y+p['sizePix']/2], linestyle='-', 
                        linewidth=p['width'], color= p['color'])
                elif p['symbol'] == "diamond":
                    c=patches.RegularPolygon([x, y], 4, radius=p['sizePix']/2, orientation=0, 
                                             color=p['color'], fill=p['fill'], linewidth=p['width'])
                    self.axes.add_patch(c)
                if l != None:
                    pylab.text(x, y+p['sizePix']/1.5, l, horizontalalignment='center', \
                                fontsize=p['objLabelSize'], color=p['color'])
            
            if p['symbol'] == "compass":
                x=p['x'][0]
                y=p['y'][0]
                ra=p['RA'][0]
                dec=p['dec'][0]
                
                westPoint,eastPoint,southPoint,northPoint=astCoords.calcRADecSearchBox(ra, dec, p['sizeArcSec']/3600.0/2.0)
                northPix=self.wcs.wcs2pix(ra, northPoint)
                eastPix=self.wcs.wcs2pix(eastPoint, dec)
                                
                edx=eastPix[0]-x
                edy=eastPix[1]-y
                ndx=northPix[0]-x
                ndy=northPix[1]-y
                nArrow=patches.Arrow(x, y, ndx, ndy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])
                eArrow=patches.Arrow(x, y, edx, edy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])                
                self.axes.add_patch(nArrow)
                self.axes.add_patch(eArrow)
                pylab.text(x+ndx+ndx*0.2, y+ndy+ndy*0.2, "N", horizontalalignment='center', 
                                verticalalignment='center', fontsize=p['objLabelSize'], color=p['color'])
                pylab.text(x+edx+edx*0.2, y+edy+edy*0.2, "E", horizontalalignment='center', 
                                verticalalignment='center', fontsize=p['objLabelSize'], color=p['color'])

            if p['symbol'] == "scaleBar":
                x=p['x'][0]
                y=p['y'][0]
                ra=p['RA'][0]
                dec=p['dec'][0]
                
                westPoint,eastPoint,southPoint,northPoint=astCoords.calcRADecSearchBox(ra, dec, p['sizeArcSec']/3600.0/2.0)
                northPix=self.wcs.wcs2pix(ra, northPoint)
                eastPix=self.wcs.wcs2pix(eastPoint, dec)
                edx=eastPix[0]-x
                edy=eastPix[1]-y
                ndx=northPix[0]-x
                ndy=northPix[1]-y
                
                if p['style'] == "arrows":
                    eArrow=patches.Arrow(x, y, edx, edy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])  
                    wArrow=patches.Arrow(x, y, -edx, edy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])   
                    self.axes.add_patch(eArrow)
                    self.axes.add_patch(wArrow)
                elif p['style'] == "whiskers":
                    ewArrow=patches.FancyArrowPatch(posA = (x+edx, y), posB = (x-edx,y+edy), edgecolor=p['color'], facecolor=p['color'], linewidth = p['width'], arrowstyle = '|-|')  
                    self.axes.add_patch(ewArrow)
                
                # Work out label
                if p['scaleBarLabel'] == None:
                    scaleLabel=None
                    if p['sizeArcSec'] < 60.0:
                        scaleLabel="%.0f %s" % (p['sizeArcSec'], DOUBLE_PRIME)
                    elif p['sizeArcSec'] >= 60.0 and p['sizeArcSec'] <  3600.0:
                        scaleLabel="%.0f %s" % (p['sizeArcSec']/60.0, PRIME)
                    else:
                        scaleLabel="%.0f %s" % (p['sizeArcSec']/3600.0, DEG)
                else:
                    scaleLabel=p['scaleBarLabel']
                pylab.text(x, y+0.025*self.data.shape[1], scaleLabel, horizontalalignment='center', 
                           verticalalignment='center', fontsize=p['objLabelSize'], color=p['color'])
                                
        if self.axesLabels != None:
            pylab.xticks(self.ticsRA[0], self.ticsRA[1], weight='normal', family=self.axesFontFamily, \
                                    fontsize=self.axesFontSize)
            pylab.yticks(self.ticsDec[0], self.ticsDec[1], weight='normal', family=self.axesFontFamily, \
                                    fontsize=self.axesFontSize)
            pylab.xlabel(self.RAAxisLabel, family=self.axesFontFamily, fontsize=self.axesFontSize)
            pylab.ylabel(self.decAxisLabel, family=self.axesFontFamily, fontsize=self.axesFontSize)
        else:
            pylab.xticks([], [])
            pylab.yticks([], [])
            pylab.xlabel("")
            pylab.ylabel("")
        
        if self.flipXAxis == False:
            pylab.xlim(0, self.data.shape[1]-1)
        else:
            pylab.xlim(self.data.shape[1]-1, 0)
        if self.flipYAxis == False:
            pylab.ylim(0, self.data.shape[0]-1)
        else:
            pylab.ylim(self.data.shape[0]-1, 0)


    def addContourOverlay(self, contourImageData, contourWCS, tag, levels = ["linear", "min", "max", 5],
                             width = 1, color = "white", smooth = 0, highAccuracy = False):
        """Adds image data to the ImagePlot as a contour overlay. The contours can be removed using 
        L{removeContourOverlay}. If a contour overlay already exists with this tag, it will be replaced.
        
        @type contourImageData: np array
        @param contourImageData: image data array from which contours are to be generated
        @type contourWCS: astWCS.WCS
        @param contourWCS: astWCS.WCS object for the image to be contoured
        @type tag: string
        @param tag: identifying tag for this set of contours
        @type levels: list
        @param levels: sets the contour levels - available options:
            - values: contourLevels=[list of values specifying each level]
            - linear spacing: contourLevels=['linear', min level value, max level value, number
            of levels] - can use "min", "max" to automatically set min, max levels from image data
            - log spacing: contourLevels=['log', min level value, max level value, number of
            levels] - can use "min", "max" to automatically set min, max levels from image data
        @type width: int
        @param width: width of the overlaid contours
        @type color: string
        @param color: color of the overlaid contours, specified by the name of a standard
            matplotlib color, e.g., "black", "white", "cyan"
            etc. (do "help(pylab.colors)" in the Python interpreter to see available options)
        @type smooth: float
        @param smooth: standard deviation (in arcsec) of Gaussian filter for
            pre-smoothing of contour image data (set to 0 for no smoothing)
        @type highAccuracy: bool
        @param highAccuracy: if True, sample every corresponding pixel in each image; otherwise, sample
            every nth pixel, where n = the ratio of the image scales.
        
        """
                
        if self.rgbImage == True:
            backgroundData=self.data[:,:,0]
        else:
            backgroundData=self.data
        contourData=astImages.generateContourOverlay(backgroundData, self.wcs, contourImageData, \
                                              contourWCS, levels, smooth, highAccuracy = highAccuracy)
        
        alreadyGot=False
        for c in self.contourOverlays:
            if c['tag'] == tag:
                c['contourData']=contourData
                c['tag']=tag
                c['color']=color
                c['width']=width
                alreadyGot=True
                
        if alreadyGot == False:
            self.contourOverlays.append({'contourData': contourData, 'tag': tag, 'color': color, \
                                            'width': width})
        self.draw()

    
    def removeContourOverlay(self, tag):
        """Removes the contourOverlay from the ImagePlot corresponding to the tag.
        
        @type tag: string
        @param tag: tag for contour overlay in ImagePlot.contourOverlays to be removed
        
        """
        
        index=0
        for p in self.contourOverlays:
            if p['tag'] == tag:
                self.plotObjects.remove(self.plotObjects[index])
            index=index+1
        self.draw()
        
        
    def addPlotObjects(self, objRAs, objDecs, tag, symbol="circle", size=4.0, width=1.0, color="yellow",\
                       fill = False, objLabels = None, objLabelSize = 12.0):
        """Add objects with RA, dec coords objRAs, objDecs to the ImagePlot. Only objects that fall within 
        the image boundaries will be plotted.
        
        symbol specifies the type of symbol with which to mark the object in the image. The following
        values are allowed:
            - "circle"
            - "box"
            - "cross"
            - "diamond"
        
        size specifies the diameter in arcsec of the symbol (if plotSymbol == "circle"), or the width
        of the box in arcsec (if plotSymbol == "box")
        
        width specifies the thickness of the symbol lines in pixels
        
        color can be any valid matplotlib color (e.g. "red", "green", etc.)
        
        The objects can be removed from the plot by using removePlotObjects(), and then calling
        draw(). If the ImagePlot already has a set of plotObjects with the same tag, they will be 
        replaced.
        
        @type objRAs: np array or list
        @param objRAs: object RA coords in decimal degrees
        @type objDecs: np array or list
        @param objDecs: corresponding object Dec. coords in decimal degrees
        @type tag: string
        @param tag: identifying tag for this set of objects
        @type symbol: string
        @param symbol: either "circle", "box", "cross", or "diamond"
        @type size: float
        @param size: size of symbols to plot (radius in arcsec, or width of box)
        @type width: float
        @param width: width of symbols in pixels
        @type color: string
        @param color: any valid matplotlib color string, e.g. "red", "green" etc.
        @type fill: bool
        @param color: if True, fill symbols
        @type objLabels: list
        @param objLabels: text labels to plot next to objects in figure
        @type objLabelSize: float
        @param objLabelSize: size of font used for object labels (in points)
        
        """
        
        pixCoords=self.wcs.wcs2pix(objRAs, objDecs)
        
        xMax=self.data.shape[1]
        yMax=self.data.shape[0]
        
        if objLabels is None:
            objLabels=[None]*len(objRAs)
            
        xInPlot=[]
        yInPlot=[]
        RAInPlot=[]
        decInPlot=[]
        labelInPlot=[]
        for p, r, d, l in zip(pixCoords, objRAs, objDecs, objLabels):
            if p[0] >= 0 and p[0] < xMax and p[1] >= 0 and p[1] < yMax:
                xInPlot.append(p[0])
                yInPlot.append(p[1])
                RAInPlot.append(r)
                decInPlot.append(d)
                labelInPlot.append(l)
        
        xInPlot=np.array(xInPlot)
        yInPlot=np.array(yInPlot)
        RAInPlot=np.array(RAInPlot)
        decInPlot=np.array(decInPlot)
        
        # Size of symbols in pixels in plot - converted from arcsec
        sizePix=(size/3600.0)/self.wcs.getPixelSizeDeg()
        
        alreadyGot=False
        for p in self.plotObjects:
            if p['tag'] == tag:
                p['x']=xInPlot
                p['y']=yInPlot
                p['RA']=RAInPlot
                p['dec']=decInPlot
                p['tag']=tag
                p['objLabels']=objLabels
                p['symbol']=symbol
                p['sizePix']=sizePix
                p['sizeArcSec']=size
                p['width']=width
                p['color']=color
                p['objLabelSize']=objLabelSize
                p['fill']=fill
                alreadyGot=True
        
        if alreadyGot == False:
            self.plotObjects.append({'x': xInPlot, 'y': yInPlot, 'RA': RAInPlot, 'dec': decInPlot,
                                'tag': tag, 'objLabels': labelInPlot, 'symbol': symbol, 
                                'sizePix': sizePix, 'width': width, 'color': color,
                                'objLabelSize': objLabelSize, 'sizeArcSec': size, 'fill': fill})
        self.draw()
        
        
    def removePlotObjects(self, tag):
        """Removes the plotObjects from the ImagePlot corresponding to the tag. The plot must be redrawn
        for the change to take effect.
        
        @type tag: string
        @param tag: tag for set of objects in ImagePlot.plotObjects to be removed
        
        """
        
        index=0
        for p in self.plotObjects:
            if p['tag'] == tag:
                self.plotObjects.remove(self.plotObjects[index])
            index=index+1
        self.draw()
  
        
    def addCompass(self, location, sizeArcSec, color = "white", fontSize = 12, \
                        width = 20.0):
        """Adds a compass to the ImagePlot at the given location ('N', 'NE', 'E', 'SE', 'S', 
        'SW', 'W', or 'NW'). Note these aren't directions on the WCS coordinate grid, they are 
        relative positions on the plot - so N is top centre, NE is top right, SW is bottom right etc.. 
        Alternatively, pixel coordinates (x, y) in the image can be given.
        
        @type location: string or tuple
        @param location: location in the plot where the compass is drawn:
            - string: N, NE, E, SE, S, SW, W or NW
            - tuple: (x, y)
        @type sizeArcSec: float
        @param sizeArcSec: length of the compass arrows on the plot in arc seconds
        @type color: string
        @param color: any valid matplotlib color string
        @type fontSize: float
        @param fontSize: size of font used to label N and E, in points
        @type width: float
        @param width: width of arrows used to mark compass
        
        """
        
        if type(location) == str:
            cRADeg, cDecDeg=self.wcs.getCentreWCSCoords()
            RAMin, RAMax, decMin, decMax=self.wcs.getImageMinMaxWCSCoords()
            westPoint,eastPoint,southPoint,northPoint=astCoords.calcRADecSearchBox(cRADeg, cDecDeg, sizeArcSec/3600.0/2.0)
            sizeRADeg=eastPoint-westPoint
            sizeDecDeg=northPoint-southPoint
            xSizePix=(sizeArcSec/3600.0)/self.wcs.getXPixelSizeDeg()
            ySizePix=(sizeArcSec/3600.0)/self.wcs.getYPixelSizeDeg()
            X=self.data.shape[1]
            Y=self.data.shape[0]
            xBufferPix=0.5*xSizePix
            yBufferPix=0.5*ySizePix
            cx, cy=self.wcs.wcs2pix(cRADeg, cDecDeg)
            foundLocation=False
            x=cy
            y=cx
            if self.wcs.isFlipped() == False:              
                if location.find("N") != -1:
                    y=Y-2*yBufferPix
                    foundLocation=True
                if location.find("S") != -1:
                    y=yBufferPix
                    foundLocation=True
                if location.find("E") != -1:
                    x=xBufferPix*2
                    foundLocation=True
                if location.find("W") != -1:
                    x=X-xBufferPix
                    foundLocation=True
            else:
                if location.find("S") != -1:
                    y=Y-2*yBufferPix
                    foundLocation=True
                if location.find("N") != -1:
                    y=yBufferPix
                    foundLocation=True
                if location.find("W") != -1:
                    x=xBufferPix*2
                    foundLocation=True
                if location.find("E") != -1:
                    x=X-xBufferPix
                    foundLocation=True
            if foundLocation == False:
                raise Exception("didn't understand location string for scale bar (should be e.g. N, S, E, W).")
            RADeg, decDeg=self.wcs.pix2wcs(x, y)
        elif type(location) == tuple or type(location) == list:
            x, y=location
            RADeg, decDeg=self.wcs.pix2wcs(x, y)
        else:
            raise Exception("didn't understand location for scale bar - should be string or tuple.")
        
        alreadyGot=False
        for p in self.plotObjects:
            if p['tag'] == "compass":
                p['x']=[x]
                p['y']=[y]
                p['RA']=[RADeg]
                p['dec']=[decDeg]
                p['tag']="compass"
                p['objLabels']=[None]
                p['symbol']="compass"
                p['sizeArcSec']=sizeArcSec
                p['width']=width
                p['color']=color
                p['objLabelSize']=fontSize
                alreadyGot=True
        
        if alreadyGot == False:
            self.plotObjects.append({'x': [x], 'y': [y], 'RA': [RADeg], 'dec': [decDeg],
                                'tag': "compass", 'objLabels': [None], 'symbol': "compass", 
                                'width': width, 'color': color,
                                'objLabelSize': fontSize, 'sizeArcSec': sizeArcSec})
        self.draw()


    def addScaleBar(self, location, sizeArcSec, color = "white", fontSize = 12, \
                        width = 20.0, label = None, style = "whiskers"):
        """Adds a scale bar to the ImagePlot at the given location ('N', 'NE', 'E', 'SE', 'S', 
        'SW', 'W', or 'NW'). Note these aren't directions on the WCS coordinate grid, they are 
        relative positions on the plot - so N is top centre, NE is top right, SW is bottom right etc.. 
        Alternatively, pixel coordinates (x, y) in the image can be given.
        
        @type location: string or tuple
        @param location: location in the plot where the compass is drawn:
            - string: N, NE, E, SE, S, SW, W or NW
            - tuple: (x, y)
        @type sizeArcSec: float
        @param sizeArcSec: scale length to indicate on the plot in arc seconds
        @type color: string
        @param color: any valid matplotlib color string
        @type fontSize: float
        @param fontSize: size of font used to label N and E, in points
        @type width: float
        @param width: width of arrow used to mark scale
        @type label: string
        @param label: overrides the displayed label if not None (if None, label is the angular size)
        @type style: string
        @param style: either "whiskers" or "arrows"
        
        """
        
        # Work out where the scale bar is going in WCS coords from the relative location given
        if type(location) == str:
            cRADeg, cDecDeg=self.wcs.getCentreWCSCoords()
            RAMin, RAMax, decMin, decMax=self.wcs.getImageMinMaxWCSCoords()
            westPoint,eastPoint,southPoint,northPoint=astCoords.calcRADecSearchBox(cRADeg, cDecDeg, sizeArcSec/3600.0/2.0)
            sizeRADeg=eastPoint-westPoint
            sizeDecDeg=northPoint-southPoint
            xSizePix=(sizeArcSec/3600.0)/self.wcs.getXPixelSizeDeg()
            ySizePix=(sizeArcSec/3600.0)/self.wcs.getYPixelSizeDeg()
            X=self.data.shape[1]
            Y=self.data.shape[0]
            xBufferPix=0.6*ySizePix
            yBufferPix=0.05*Y
            cx, cy=self.wcs.wcs2pix(cRADeg, cDecDeg)
            foundLocation=False
            x=cy
            y=cx
            if self.wcs.isFlipped() == False:
                if location.find("N") != -1:
                    y=Y-1.5*yBufferPix
                    foundLocation=True
                if location.find("S") != -1:
                    y=yBufferPix
                    foundLocation=True
                if location.find("E") != -1:
                    x=xBufferPix
                    foundLocation=True
                if location.find("W") != -1:
                    x=X-xBufferPix
                    foundLocation=True
            else:
                if location.find("S") != -1:
                    y=Y-1.5*yBufferPix
                    foundLocation=True
                if location.find("N") != -1:
                    y=yBufferPix
                    foundLocation=True
                if location.find("W") != -1:
                    x=xBufferPix
                    foundLocation=True
                if location.find("E") != -1:
                    x=X-xBufferPix
                    foundLocation=True
            if foundLocation == False:
                raise Exception("didn't understand location string for scale bar (should be e.g. N, S, E, W).")
            RADeg, decDeg=self.wcs.pix2wcs(x, y)
        elif type(location) == tuple or type(location) == list:
            x, y=location
            RADeg, decDeg=self.wcs.pix2wcs(x, y)
        else:
            raise Exception("didn't understand location for scale bar - should be string or tuple.")
        
        alreadyGot=False
        for p in self.plotObjects:
            if p['tag'] == "scaleBar":
                p['x']=[x]
                p['y']=[y]
                p['RA']=[RADeg]
                p['dec']=[decDeg]
                p['tag']="scaleBar"
                p['objLabels']=[None]
                p['symbol']="scaleBar"
                p['sizeArcSec']=sizeArcSec
                p['width']=width
                p['color']=color
                p['objLabelSize']=fontSize
                p['scaleBarLabel']=label
                p['style']=style
                alreadyGot=True
        
        if alreadyGot == False:
            self.plotObjects.append({'x': [x], 'y': [y], 'RA': [RADeg], 'dec': [decDeg],
                                'tag': "scaleBar", 'objLabels': [None], 'symbol': "scaleBar", 
                                'width': width, 'color': color,
                                'objLabelSize': fontSize, 'sizeArcSec': sizeArcSec,
                                'scaleBarLabel': label, 'style': style})
        self.draw()
                                

    def calcWCSAxisLabels(self, axesLabels = "decimal"):
        """This function calculates the positions of coordinate labels for the RA and Dec axes of the 
        ImagePlot. The tick steps are calculated automatically unless self.RATickSteps,
        self.decTickSteps are set to values other than "auto" (see L{ImagePlot.__init__}). 
        
        The ImagePlot must be redrawn for changes to be applied.
        
        @type axesLabels: string
        @param axesLabels: either "sexagesimal" (for H:M:S, D:M:S), "decimal" (for decimal degrees),
        or None for no coordinate axes labels
        
        """
        
        # Label equinox on axes
        equinox=self.wcs.getEquinox()
        if equinox<1984:
            equinoxLabel="B"+str(int(equinox))
        else:
            equinoxLabel="J"+str(int(equinox))
           
        self.axesLabels=axesLabels
        
        ticsDict=self.getTickSteps()
        
        # Manual override - note: no minor tick marks anymore, but may want to bring them back
        if self.RATickSteps != "auto":
            ticsDict['major']['RA']=self.RATickSteps
        if self.decTickSteps != "auto":
            ticsDict['major']['dec']=self.decTickSteps
        
        RALocs=[]
        decLocs=[]
        RALabels=[]
        decLabels=[]
        key="major"
        #for key in ticsDict.keys(): # key is major or minor
        if self.axesLabels == "sexagesimal":
            self.RAAxisLabel="R.A. ("+equinoxLabel+")"
            self.decAxisLabel="Dec. ("+equinoxLabel+")"
            RADegStep=ticsDict[key]['RA']['deg']
            decDegStep=ticsDict[key]['dec']['deg']
        elif self.axesLabels == "decimal":
            self.RAAxisLabel="R.A. Degrees ("+equinoxLabel+")"
            self.decAxisLabel="Dec. Degrees ("+equinoxLabel+")"
            RADegStep=ticsDict[key]['RA']
            decDegStep=ticsDict[key]['dec']
        else:
            raise Exception("axesLabels must be either 'sexagesimal' or 'decimal'")
        
        # xArray=np.arange(0, self.data.shape[1], 1)
        xArray=np.linspace(0, self.data.shape[1], self.data.shape[1]*2)
        yArray=np.arange(0, self.data.shape[0], 1)
        xWCS=self.wcs.pix2wcs(xArray, np.zeros(xArray.shape[0], dtype=float))
        yWCS=self.wcs.pix2wcs(np.zeros(yArray.shape[0], dtype=float), yArray)
        xWCS=np.array(xWCS)
        yWCS=np.array(yWCS)
        ras=xWCS[:,0]
        # ras[ras < 0]=ras[ras < 0]+360
        decs=yWCS[:,1]
        RAEdges=np.array([ras[0], ras[-1]])
        # RAMin=RAEdges.min()
        # RAMax=RAEdges.max()
        RAMin=ras.min()
        RAMax=ras.max()
        decMin=decs.min()
        decMax=decs.max()
        
        # Work out if wrapped around
        midRAPix, midDecPix=self.wcs.wcs2pix((RAEdges[1]+RAEdges[0])/2.0, (decMax+decMin)/2.0)
        if midRAPix < 0 or midRAPix > self.wcs.header['NAXIS1']:
            wrappedRA=True
        else:
            wrappedRA=False
            
        # Note RA, dec work in opposite sense below because E at left
        if ras[1] < ras[0]:
            self.flipXAxis=False
            ra2x=interpolate.interp1d(ras[::-1], xArray[::-1], kind='linear')
        else:
            self.flipXAxis=True
            ra2x=interpolate.interp1d(ras, xArray, kind='linear')
        if decs[1] < decs[0]:
            self.flipYAxis=True
            dec2y=interpolate.interp1d(decs[::-1], yArray[::-1], kind='linear')
        else:
            self.flipYAxis=False
            dec2y=interpolate.interp1d(decs, yArray, kind='linear')
        
        if wrappedRA == False:
            RAPlotMin=RADegStep*math.modf(RAMin/RADegStep)[1]
            RAPlotMax=RADegStep*math.modf(RAMax/RADegStep)[1]
            if RAPlotMin < RAMin:
                RAPlotMin=RAPlotMin+RADegStep
            if RAPlotMax >= RAMax:
                RAPlotMax=RAPlotMax-RADegStep
            RADegs=np.arange(RAPlotMin, RAPlotMax+0.0001, RADegStep)
        else:
            RAPlotMin=RADegStep*math.modf(RAMin/RADegStep)[1]
            RAPlotMax=RADegStep*math.modf(RAMax/RADegStep)[1]
            if RAPlotMin > RAMin:
                RAPlotMin=RAPlotMin-RADegStep
            if RAPlotMax <= RAMax:
                RAPlotMax=RAPlotMax+RADegStep
            for i in range(ras.shape[0]):
                if ras[i] >= RAMax and ras[i] <= 360.0:
                    ras[i]=ras[i]-360.0
            if ras[1] < ras[0]:
                ra2x=interpolate.interp1d(ras[::-1], xArray[::-1], kind='linear')
            else:
                ra2x=interpolate.interp1d(ras, xArray, kind='linear')
            RADegs=np.arange(RAPlotMin, RAPlotMax-360.0-0.0001, -RADegStep)

        # Full RA range?
        if round(self.wcs.getXPixelSizeDeg()*self.wcs.header['NAXIS1']) == 360:
            RADegs=np.arange(0, 360, RADegStep)

        decPlotMin=decDegStep*math.modf(decMin/decDegStep)[1]
        decPlotMax=decDegStep*math.modf(decMax/decDegStep)[1]
        if decPlotMin < decMin:
            decPlotMin=decPlotMin+decDegStep
        if decPlotMax >= decMax:
            decPlotMax=decPlotMax-decDegStep
        decDegs=np.arange(decPlotMin, decPlotMax+0.0001, decDegStep)
        
        if key == "major":
            if axesLabels == "sexagesimal":
                for r in RADegs:
                    if r < 0:
                        r=r+360.0
                    h, m, s=astCoords.decimal2hms(r, ":").split(":")
                    hInt=int(round(float(h)))
                    if ticsDict[key]['RA']['unit'] == 'h' and (60.0-float(m)) < 0.01: # Check for rounding error
                        hInt=hInt+1
                    if hInt < 10:
                        hString="0"+str(hInt)
                    else:
                        hString=str(hInt)
                    mInt=int(round(float(m)))
                    if ticsDict[key]['RA']['unit'] == 'm' and (60.0-float(s)) < 0.01: # Check for rounding error
                        mInt=mInt+1
                    if mInt < 10:
                        mString="0"+str(mInt)
                    else:
                        mString=str(mInt)
                    sInt=int(round(float(s)))
                    if sInt < 10:
                        sString="0"+str(sInt)
                    else:
                        sString=str(sInt)
                    if ticsDict[key]['RA']['unit'] == 'h':
                        rString=hString+"$^{\sf{h}}$"
                    elif ticsDict[key]['RA']['unit'] == 'm':
                        rString=hString+"$^{\sf{h}}$"+mString+"$^{\sf{m}}$"
                    else:
                        rString=hString+"$^{\sf{h}}$"+mString+"$^{\sf{m}}$"+sString+"$^{\sf{s}}$"
                    RALabels.append(rString)
                for D in decDegs:
                    d, m, s=astCoords.decimal2dms(D, ":").split(":")
                    dInt=int(round(float(d)))
                    if ticsDict[key]['dec']['unit'] == 'd' and (60.0-float(m)) < 0.01: # Check for rounding error
                        dInt=dInt+1
                    if dInt < 10 and dInt >= 0 and D > 0:
                        dString="+0"+str(dInt)
                    elif dInt > -10 and dInt <= 0 and D < 0:
                        dString="-0"+str(abs(dInt))
                    elif dInt >= 10:
                        dString="+"+str(dInt)
                    else:
                        dString=str(dInt)
                    mInt=int(round(float(m)))
                    if ticsDict[key]['dec']['unit'] == 'm' and (60.0-float(s)) < 0.01: # Check for rounding error
                        mInt=mInt+1
                    if mInt < 10:
                        mString="0"+str(mInt)
                    else:
                        mString=str(mInt)
                    sInt=int(round(float(s)))
                    if sInt < 10:
                        sString="0"+str(sInt)
                    else:
                        sString=str(sInt)
                    if ticsDict[key]['dec']['unit'] == 'd':
                        dString=dString+DEG
                    elif ticsDict[key]['dec']['unit'] == 'm':
                        dString=dString+DEG+mString+PRIME
                    else:
                        dString=dString+DEG+mString+PRIME+sString+DOUBLE_PRIME               
                    decLabels.append(dString)
            elif axesLabels == "decimal":
                                
                if wrappedRA == False:
                    RALabels=RALabels+RADegs.tolist()
                else:
                    nonNegativeLabels=[]
                    for r in RADegs:
                        if r < 0:
                            r=r+360.0
                        nonNegativeLabels.append(r)
                    RALabels=RALabels+nonNegativeLabels
                decLabels=decLabels+decDegs.tolist()
                
                # Format RALabels, decLabels to same number of d.p.
                dpNumRA=len(str(ticsDict['major']['RA']).split(".")[-1])
                dpNumDec=len(str(ticsDict['major']['dec']).split(".")[-1])
                for i in range(len(RALabels)):
                    fString="%."+str(dpNumRA)+"f"
                    RALabels[i]=fString % (RALabels[i])
                for i in range(len(decLabels)):
                    fString="%."+str(dpNumDec)+"f"
                    decLabels[i]=fString % (decLabels[i])                                
        
        if key == 'minor':
            RALabels=RALabels+RADegs.shape[0]*['']
            decLabels=decLabels+decDegs.shape[0]*['']

        decLocs=decLocs+dec2y(decDegs).tolist()

        # This is for avoiding 0h <-> 12h degeneracy that sometimes happens
        RALocs=[]
        RALocs=RALocs+ra2x(RADegs).tolist()
        uniqRALocs=[]
        uniqRALabels=[]
        for i in range(len(RALocs)):
            if RALocs[i] not in uniqRALocs:
                uniqRALocs.append(RALocs[i])
                uniqRALabels.append(RALabels[i])
        RALocs=uniqRALocs
        RALabels=uniqRALabels

        self.ticsRA=[RALocs, RALabels]
        self.ticsDec=[decLocs, decLabels]
        

    def save(self, fileName):
        """Saves the ImagePlot in any format that matplotlib can understand, as determined from the 
        fileName extension.
        
        @type fileName: string
        @param fileName: path where plot will be written
        
        """
        
        pylab.draw()
        pylab.savefig(fileName)
        
        
    def getTickSteps(self):
        """Chooses the appropriate WCS coordinate tick steps for the plot based on its size.
        Whether the ticks are decimal or sexagesimal is set by self.axesLabels.
        
        Note: minor ticks not used at the moment.
        
        @rtype: dictionary
        @return: tick step sizes for major, minor plot ticks, in format {'major', 'minor'}
                
        """
        
        # Aim for 5 major tick marks on a plot
        xArray=np.arange(0, self.data.shape[1], 1)
        yArray=np.arange(0, self.data.shape[0], 1)
        xWCS=self.wcs.pix2wcs(xArray, np.zeros(xArray.shape[0], dtype=float))
        yWCS=self.wcs.pix2wcs(np.zeros(yArray.shape[0], dtype=float), yArray)
        xWCS=np.array(xWCS)
        yWCS=np.array(yWCS)
        ras=xWCS[:,0]
        decs=yWCS[:,1]
        RAEdges=np.array([ras[0], ras[-1]])
        RAMin=RAEdges.min()
        RAMax=RAEdges.max()
        decMin=decs.min()
        decMax=decs.max()
        
        # Work out if wrapped around
        midRAPix, midDecPix=self.wcs.wcs2pix((RAEdges[1]+RAEdges[0])/2.0, (decMax+decMin)/2.0)
        if midRAPix < 0 or midRAPix > self.wcs.header['NAXIS1']:
            wrappedRA=True
        else:
            wrappedRA=False
        if wrappedRA == False:
            RAWidthDeg=RAMax-RAMin
        else:
            RAWidthDeg=(360.0-RAMax)+RAMin
        decHeightDeg=decMax-decMin

        ticsDict={}
        ticsDict['major']={}
        ticsDict['minor']={}
        if self.axesLabels == "sexagesimal":
            
            matchIndex = 0
            for i in range(len(RA_TICK_STEPS)):
                if RAWidthDeg/2.5 > RA_TICK_STEPS[i]['deg']:
                    matchIndex = i
            
            ticsDict['major']['RA']=RA_TICK_STEPS[matchIndex]
            ticsDict['minor']['RA']=RA_TICK_STEPS[matchIndex-1]

            matchIndex = 0
            for i in range(len(DEC_TICK_STEPS)):
                if decHeightDeg/2.5 > DEC_TICK_STEPS[i]['deg']:
                    matchIndex = i
                                
            ticsDict['major']['dec']=DEC_TICK_STEPS[matchIndex]
            ticsDict['minor']['dec']=DEC_TICK_STEPS[matchIndex-1]
            
            return ticsDict
            
        elif self.axesLabels == "decimal":
            
            matchIndex = 0
            for i in range(len(DECIMAL_TICK_STEPS)):
                if RAWidthDeg/2.5 > DECIMAL_TICK_STEPS[i]:
                    matchIndex = i
            
            ticsDict['major']['RA']=DECIMAL_TICK_STEPS[matchIndex]
            ticsDict['minor']['RA']=DECIMAL_TICK_STEPS[matchIndex-1]
            
            matchIndex = 0
            for i in range(len(DECIMAL_TICK_STEPS)):
                if decHeightDeg/2.5 > DECIMAL_TICK_STEPS[i]:
                    matchIndex = i
                    
            ticsDict['major']['dec']=DECIMAL_TICK_STEPS[matchIndex]
            ticsDict['minor']['dec']=DECIMAL_TICK_STEPS[matchIndex-1]
            
            return ticsDict
        
        else:
            raise Exception("axesLabels must be either 'sexagesimal' or 'decimal'")




#!/usr/bin/env python
#File: simpleRGB/simpleRGB.py
#Created: Sat Dec 15 17:25:44 2012
#Last Change: Sat Dec 15 17:26:38 2012
# -*- coding: utf-8 -*-
#
# Simple RGB ImagePlot example

import numpy
import astropy.io.fits as pyfits
import pylab
from astLib import *

# Load the images - these have to be aligned to pixel and same pixel
# dimensions
rimg = pyfits.open("../../../testingData/stephanDSS2IR.fits")
gimg = pyfits.open("../../../testingData/stephanDSS2Red.fits")
bimg = pyfits.open("../../../testingData/stephanDSS2Blue.fits")
wcs = astWCS.WCS("../../../testingData/stephanDSS2Blue.fits")
r = rimg[0].data
g = gimg[0].data
b = bimg[0].data
rCut = [r.min(), r.max()]
gCut = [g.min(), g.max()]
bCut = [b.min(), b.max()]

# Make the figure
p = astPlots.ImagePlot([r, g, b], wcs, axes = [0.12, 0.12, 0.8, 0.8], cutLevels = [rCut, gCut, bCut],
    title="Stephan's Quintet")
p.save("output_simpleRGB.png")



#!/usr/bin/env python
#File: CMRRestFrameConversion.py
#Created: Sat Dec 15 17:03:04 2012
#Last Change: Sat Dec 15 17:03:50 2012
# -*- coding: utf-8 -*-
#
# Calculates (U-V)z slopes, scatters and intercepts of a given color--magnitude
# relation.
# Follows the procedure in Appendix II of Mei et al. 2009 (ApJ, 690, 42),
# except here we convert
# mags to apparent mags at the distance of the Coma cluster.

from astLib import astSED
from astLib import astStats
from astLib import astCalc
from scipy import stats
from scipy import optimize
from scipy import interpolate
import numpy
import pylab
import os
import sys
import random
import math
import pickle
import string
random.seed()

#-----------------------------------------------------------------------------
# Constants etc.

# number of bootstrap samples, for estimating errors
BOOTSTRAPS = 1000
# number of galaxies, gets * number of models, having same age in each
# metallicity bin
NGAL = 25
FILTER_DIR = "../../../testingData/filters/"

# Map between short filter names on command line and paths, labels etc..
filterMap=[]
filterMap.append({'shortName': 'r625', 'filePath': FILTER_DIR+'F625W_WFC.res',
'plotLabel': 'r625'})
filterMap.append({'shortName': 'i775', 'filePath': FILTER_DIR+'F775W_WFC.res',
'plotLabel': 'i775'})
filterMap.append({'shortName': 'z850', 'filePath': FILTER_DIR+'F850LP_WFC.res',
'plotLabel': 'z850'})
filterMap.append({'shortName': 'U', 'filePath': FILTER_DIR+'U_Johnson.res',
'plotLabel': 'U'})
filterMap.append({'shortName': 'V', 'filePath': FILTER_DIR+'V_Johnson.res',
'plotLabel': 'V'})

# Literature CMR results we want to convert, in their native format
litCMRs = []
litCMRs.append({'name': 'RX J0152.7-1357 (Mei et al. 2009)',
                'redshift': 0.83,
                'colour': 'r625-z850',
                'mag': 'i775',
                'slope': -0.040,
                'slopeErr': 0.017,
                'intercept': 1.93,
                'interceptErr': 0.02,
                'zeroMag': 22.5,
                'scatter': 0.079,
                'scatterErr': 0.008,
                'magType': "AB"})

#-----------------------------------------------------------------------------
def GetCMR(nameFragment, dictList):
    """Finds the CMR dictionary in the litCMRs or results list, by looking for
    nameFragment in name.

    """

    foundCMR = None

    for dict in dictList:
        if nameFragment in dict['name']:
            foundCMR = dict

    return foundCMR

#------------------------------------------------------------------------------
def CalcTransformedCMRWithZM(obsCMR, fitMags, fitCols):
    """Calculates the CMR transformed to the rest frame, using the results of
    the magnitude and colour
    conversion fits. See handwritten notes for the tedious algebra involved.

    """

    # we use the following notation s, zp for slope, zeropoint, append Err for
    # errors
    # CMR for CMR, Mag for fitMags, Col for fitCols
    sCMR = obsCMR['slope']
    zpCMR = obsCMR['intercept']
    zmCMR = obsCMR['zeroMag']

    sMag = fitMags['slope']
    zpMag=fitMags['intercept']

    sCol = fitCols['slope']
    zpCol = fitCols['intercept']

    transformedZeroMag = obsCMR['transformedZeroMag']

    a = sCMR**-1+sMag
    b = sCol/a
    c = zpCMR/sCMR
    d = c-zpMag-zmCMR
    e = d/a
    f = sCol*e
    # this last term is if we want to transform to e.g. match Mei et al.
    g = zpCol+f+(b*transformedZeroMag)

    restCMRSlope = b
    restCMRIntercept = g
    restCMRScatter = obsCMR['scatter']*fitCols['slope']

    return ({'slope': restCMRSlope, 'intercept': restCMRIntercept,
                'scatter': restCMRScatter})

#-----------------------------------------------------------------------------
def BootstrapTransformedCMRErrorsWithZM(obsCMR, fitMags, fitCols):
    """Estimates errors on transformed CMR fit (i.e., into rest frame), by
    assuming the errors on the observed CMR fit, colour transformation fit, and
    mag. transformation fit have Gaussian
    distributions.

    """

    # we use the following notation s, zp for slope, zeropoint, append Err for
    # errors
    # CMR for CMR, Mag for fitMags, Col for fitCols
    sCMR = obsCMR['slope']
    sCMRErr = obsCMR['slopeErr']
    zpCMR = obsCMR['intercept']
    zpCMRErr = obsCMR['interceptErr']

    sMag = fitMags['slope']
    sMagErr = fitMags['slopeError']
    zpMag = fitMags['intercept']
    zpMagErr = fitMags['interceptError']

    sCol = fitCols['slope']
    sColErr = fitCols['slopeError']
    zpCol = fitCols['intercept']
    zpColErr = fitCols['interceptError']

    bsFitResults = []
    for n in range(BOOTSTRAPS):

        bsCMR = {}
        bsMag = {}
        bsCol = {}

        bsCMR['slope'] = random.normalvariate(sCMR, sCMRErr)
        bsCMR['intercept'] = random.normalvariate(zpCMR, zpCMRErr)
        bsCMR['zeroMag'] = obsCMR['zeroMag']
        bsMag['slope'] = random.normalvariate(sMag, sMagErr)
        bsMag['intercept'] = random.normalvariate(zpMag, zpMagErr)
        bsCol['slope'] = random.normalvariate(sCol, sColErr)
        bsCol['intercept'] = random.normalvariate(zpCol, zpColErr)

        bsCMR['scatter'] = random.normalvariate(obsCMR['scatter'],
                        obsCMR['scatterErr'])

        bsCMR['transformedZeroMag'] = obsCMR['transformedZeroMag']

        bsFitResults.append(CalcTransformedCMRWithZM(bsCMR, bsMag, bsCol))

    bsSlopes = []
    bsIntercepts = []
    bsScatters = []
    for bsResult in bsFitResults:
        bsSlopes.append(bsResult['slope'])
        bsIntercepts.append(bsResult['intercept'])
        bsScatters.append(bsResult['scatter'])

    bsSlopes = numpy.array(bsSlopes)
    bsIntercepts = numpy.array(bsIntercepts)
    bsScatters = numpy.array(bsScatters)

    restCMRSlopeErr = numpy.std(bsSlopes)
    restCMRInterceptErr = numpy.std(bsIntercepts)
    restCMRScatterErr = numpy.std(bsScatters)

    return ({'slopeErr': restCMRSlopeErr, 'interceptErr':restCMRInterceptErr,
                'scatterErr': restCMRScatterErr})

#-----------------------------------------------------------------------------
def LoadModels(fileNameList, modelType = "bc03"):
    """Creates a list of stellar population models from the given list of model
    file names.

    """

    models = []
    for f in fileNameList:
        if modelType == "bc03":
            models.append(astSED.BC03Model(f))
        elif modelType == "m05":
            models.append(astSED.M05Model(f))

    return models

#-----------------------------------------------------------------------------
def GetPassbandFileNames(inputColour):
    """Given a mag (e.g. i775) or colour string e.g. r625-z850, lookup the
    appropriate file name(s) in the filterMap, and return the paths in a list.

    """

    bands = inputColour.split("-")
    p = []
    for b in bands:
        p.append(None)
    for row in filterMap:
        for i in range(len(bands)):
            if row['shortName'] == bands[i]:
                p[i]=row['filePath']

    if None in p:
        print("ERROR : couldn't parse colour using filterMap")
        sys.exit()
    else:
        return p

#-----------------------------------------------------------------------------
def LoadPassbands(fileNameList, redshift = None, redshiftPassbands = False):
    """Creates a list of passband objects from the given list of passband file
    names.

    """

    passbands = []
    for f in fileNameList:
        p = astSED.Passband(f)
        if redshiftPassbands == True and redshift != None:
            p.wavelength=p.wavelength*(1.0+redshift)
        passbands.append(p)

    return passbands

#-----------------------------------------------------------------------------
def CalcColourMagTransformation(cmr, restColPassbands, restMagPassband):
    """Calculates the transformation equations needed to convert the given cmr
    into the rest frame at Coma, for the given passbands.

    """

    print((">>> Calculating colour, mag transform for CMR " +
        litCMR['name']+"..."))

    inputColour = cmr['colour']
    inputMag = cmr['mag']
    zCluster = cmr['redshift']

    # Range of formation zs to match Mei et al. 2008
    zfMax = 7.0
    zfMin = 2.0

    observedColPassbandFileNames = GetPassbandFileNames(inputColour)
    observedMagPassbandFileName = GetPassbandFileNames(inputMag)

    observedColLabel = inputColour
    observedMagLabel = inputMag

    # Load stuff
    observedColPassbands = LoadPassbands(observedColPassbandFileNames)
    observedMagPassband = LoadPassbands(observedMagPassbandFileName)[0]

    # Generate galaxy models, we'll hold them all in memory here and use them
    # all in a bit
    print("--> Generating simulated galaxy sample ...")
    restGalaxies = []
    observedGalaxies = []
    for n in range(NGAL):

        print("... n = "+str(n+1)+"/"+str(NGAL)+" ...")

        zfChoice = random.uniform(zfMin, zfMax)
        ageChoice = astCalc.tl(zfChoice)-astCalc.tl(zCluster)

        for i in range(len(models)):

            modelChoice = i
            restGalaxies.append(models[modelChoice].getSED(ageChoice, z=0.02))
            observedGalaxies.append(models[modelChoice].getSED(ageChoice,
                z=zCluster))

    # Fit for colour conversion
    observedColours = []
    restColours = []
    for o, r in zip(observedGalaxies, restGalaxies):
        restColours.append(r.calcColour(restColPassbands[0],
            restColPassbands[1], magType="Vega"))
        observedColours.append(o.calcColour(observedColPassbands[0],
            observedColPassbands[1], magType=cmr['magType']))

    restColours = numpy.array(restColours)
    observedColours = numpy.array(observedColours)

    fitColData = []
    for x, y in zip(observedColours, restColours):
        fitColData.append([x, y])

    fitCols=astStats.OLSFit(fitColData)

    res = restColours-(fitCols['slope']*observedColours+fitCols['intercept'])
    # scatter of residuals, use as fit error
    scatter = astStats.biweightScale(res, 6.0)

    # Fit for mag conversion
    restMinusObservedAppMags = []
    for o, r, obsCol, restCol in zip(observedGalaxies, restGalaxies,
        observedColours, restColours):

        restMinusObservedAppMags.append(r.calcMag(restMagPassband,
            magType="Vega")-o.calcMag(observedMagPassband,
            magType=cmr['magType']))

    restMinusObservedAppMags = numpy.array(restMinusObservedAppMags)

    fitMagData = []
    for x, y in zip(observedColours, restMinusObservedAppMags):
        fitMagData.append([x, y])

    fitMags = astStats.OLSFit(fitMagData)

    trans = {'name': cmr['name'],
            'fitCols': fitCols,
            'fitMags': fitMags,
            'observedColours': observedColours,
            'restColours': restColours,
            'restMinusObservedAppMags': restMinusObservedAppMags}

    return trans

#-----------------------------------------------------------------------------
def GetCSPModel(labelToFind, models, modelLabels):
    """Given a list of models and a matching list of labels, returns the model
    matching the given label.

    """

    foundModel = None
    for m, l in zip(models, modelLabels):
        if l == labelToFind:
            foundModel = m

    return foundModel

#-----------------------------------------------------------------------------
def ApplyCMRTransformation(cmr, trans):
    """Applies the magnitude and colour transformations stored in trans to the
CMR.

    """

    print((">>> Transforming CMR "+cmr['name']+" to Coma rest frame"+
        restColour+" ..."))

    # Check that col, mag transformations and cmrs match up, otherwise
    # something is seriously screwed up
    if cmr['name'] != trans['name']:
        print("ERROR: cmrs and transformation lists not paired!")
        sys.exit()

    fitMags = trans['fitMags']
    fitCols = trans['fitCols']
    transformedCMR = CalcTransformedCMRWithZM(cmr, fitMags, fitCols)
    transformedCMRErrs = BootstrapTransformedCMRErrorsWithZM(cmr, fitMags,
                        fitCols)

    result = {'name': cmr['name'],
            'redshift': cmr['redshift'],
            'colour': restColour,
            'mag': restMag,
            'slope': transformedCMR['slope'],
            'slopeErr': transformedCMRErrs['slopeErr'],
            'intercept': transformedCMR['intercept'],
            'interceptErr': transformedCMRErrs['interceptErr'],
            'zeroMag': cmr['transformedZeroMag'],
            'scatter': transformedCMR['scatter'],
            'scatterErr': transformedCMRErrs['scatterErr']}

    return result

#-----------------------------------------------------------------------------
# Main

# Input parameters
MODEL_TYPE="bc03"

modelFileNames = ["../../../testingData/models/tau0p1Gyr_m42.20",
                "../../../testingData/models/tau0p1Gyr_m52.20",
                "../../../testingData/models/tau0p1Gyr_m62.20",
                "../../../testingData/models/tau0p1Gyr_m72.20"]
models = LoadModels(modelFileNames, modelType = MODEL_TYPE)

# Target colour and mag bands
restColPassbandFileNames = [FILTER_DIR+"U_Johnson.res",
                        FILTER_DIR+"B_Johnson.res"]
restMagPassbandFileName = [FILTER_DIR+"B_Johnson.res"]
restColour = "U-B"    # component of output file name
restMag = "B"
restColLabel = "(U-B)rest"
restMagLabel = "B"
restColPassbands = LoadPassbands(restColPassbandFileNames)
restMagPassband = LoadPassbands(restMagPassbandFileName)[0]

# Evaluate CMR zero point in the rest frame of Coma at intercept of zero
for litCMR in litCMRs:
    litCMR['transformedZeroMag'] = 0.0

# Calculate the colour, mag transformations to take each CMR to the Coma rest frame
transformations = []
for litCMR in litCMRs:
    trans = CalcColourMagTransformation(litCMR, restColPassbands,
            restMagPassband)
    transformations.append(trans)

# Transform the literature CMRs to the rest frame passbands at Coma
results = []
for litCMR, trans in zip(litCMRs, transformations):
    result = ApplyCMRTransformation(litCMR, trans)
    results.append(result)

# Write results to a text file
outFile = open("output_CMRRestConversion.txt", "w")

# Colour, mag transformation
outFile.write("# Color, mag %s rest frame transformation fit coeffs:\n" %
            (restColour))
for t in transformations:
    outFile.write("name = %s\n" % (t['name']))
    outFile.write("# Color transformation:\n")
    outFile.write("slope = %.5f\n" % (t['fitCols']['slope']))
    outFile.write("slopeError = %.5f\n" % (t['fitCols']['slopeError']))
    outFile.write("intercept = %.5f\n" % (t['fitCols']['intercept']))
    outFile.write("interceptError = %.5f\n" % (t['fitCols']['interceptError']))
    outFile.write("# Mag transformation:\n")
    outFile.write("slope = %.5f\n" % (t['fitMags']['slope']))
    outFile.write("slopeError = %.5f\n" % (t['fitMags']['slopeError']))
    outFile.write("intercept = %.5f\n" % (t['fitMags']['intercept']))
    outFile.write("interceptError = %.5f\n" % (t['fitMags']['interceptError']))

# Transformed CMR
outFile.write("# Transformed CMR:\n")
keyOrder = ["name", "redshift", "colour", "mag", "zeroMag", "slope",
            "slopeErr", "intercept", "interceptErr", "scatter", "scatterErr"]
for r in results:
    for k in keyOrder:
        for key in list(r.keys()):
            if str(key) == k:
                if type(r[key]) == str:
                    outFile.write("%s = %s\n" % (key, r[key]))
                else:
                    outFile.write("%s = %.3f\n" % (key, float(r[key])))
outFile.close()



"""
astLib - python astronomy modules

(c) 2007-2012 Matt Hilton

(c) 2013-2025 Matt Hilton & Steven Boada

U{https://astlib.readthedocs.io}

See the README file for information on usage and installation.
See the LICENSE file for information on distribution.

"""

__all__=['astCalc', 'astCoords', 'astImages', 'astPlots', 'astStats', 'astWCS', 'astSED']
__version__ = '0.13.2'



#!/usr/bin/env python
#File: rgbTest/rgbTest.py
#Created: Sat Dec 15 17:29:50 2012
#Last Change: Sat Dec 15 17:35:08 2012
# -*- coding: utf-8 -*-
#
# Example RGB ImagePlot with objects marked and a contour overlay

import numpy
import astropy.io.fits as pyfits
import pylab
from astLib import *

# Load in list of objects to plot - a tab-delimited text file in format id, ra, dec, group
sRAs = []
sDecs = []
sLabels = []
pRAs = []
pDecs = []
pLabels = []
inFile = open("../../../testingData/testData.csv", "r")
lines = inFile.readlines()
for line in lines:
    if line[0] != "#":
        bits = line.split()
        label = bits[0]
        ra = float(bits[1])
        dec = float(bits[2])
        group = int(bits[3])
        if group == 1:
            sLabels.append(label)
            sRAs.append(ra)
            sDecs.append(dec)
        elif group == 2:
            pLabels.append(label)
            pRAs.append(ra)
            pDecs.append(dec)
inFile.close()

# Load the images - these have to be aligned to pixel and same pixel dimensions
rimg = pyfits.open("../../../testingData/testImageR.fits")
gimg = pyfits.open("../../../testingData/testImageG.fits")
bimg = pyfits.open("../../../testingData/testImageB.fits")
wcs = astWCS.WCS("../../../testingData/testImageR.fits")
r = rimg[0].data
g = gimg[0].data
b = bimg[0].data
rCut = [-50, 1000]
gCut = [-50, 1000]
bCut = [-50, 800]

ximg = pyfits.open("../../../testingData/testXRayImage.fits")
xwcs = astWCS.WCS("../../../testingData/testXRayImage.fits")
x = ximg[0].data
cLevels = [2e-05, 2.517e-05, 3.16764e-05, 3.98647e-05, 5.01697e-05,
            6.31385e-05, 7.94597e-05, 0.0001]

# Make the figure - this is a big, high-res version
pylab.figure(figsize=(20,20))
p = astPlots.ImagePlot([r, g, b], wcs, cutLevels = [rCut, gCut, bCut],
    axesLabels="sexagesimal", axesFontSize=26.0, axes = [0.12,0.12,0.8,0.8])

p.addContourOverlay(x, xwcs, 'xmm-contours', levels=cLevels, smooth=5.0)
p.addPlotObjects(sRAs, sDecs, 'spec-members', objLabels=sLabels,
    symbol="circle", color="cyan", width=2.0, size=2.0, objLabelSize=12.0)
p.addPlotObjects(pRAs, pDecs, 'phot-members', objLabels=pLabels, symbol="box",
    color="yellow", width=2.0, size=2.0, objLabelSize=12.0)

# Add a compass because we can
p.addCompass("SW", 10, width=50.0, fontSize=30.0)

p.draw()
p.save("output_rgbTest.png")



#!/usr/bin/python
#File: coordConvert/coordConvert.py
#Created: Sat Dec 15 17:08:26 2012
#Last Change: Sat Dec 15 17:15:07 2012
# -*- coding: utf-8 -*-
#
# Converts file fed in into hms, dms format
#

from astLib import astCoords
import sys

# Main
if len(sys.argv) < 8:
    print ("Run: % coordConvert.py <in file> <out file> <RA column> <dec column> <way [either \"to_decimal\" or \"to_hmsdms\"]> <coord delimiter> <output column delimiter [e.g. \"tab\" \",\"\"&\" etc.]>")
else:

    inFile = sys.argv[1]
    outFile = sys.argv[2]
    RACol = int(sys.argv[3])-1
    decCol = int(sys.argv[4])-1
    way = sys.argv[5]
    delimiter = sys.argv[6]
    colDelim = str(sys.argv[7]).rstrip("\n")

    if colDelim == "tab":
        colDelim = "\t"

    wayOk = False
    if way == "to_decimal":
        wayOk = True
    if way == "to_hmsdms":
        wayOk = True
    if wayOk == False:
        print("<way>: either \"to_decimal\" or \"to_hmsdms\"")
        sys.exit()

    reader = open(inFile, "r")
    lines = reader.readlines()
    reader.close()

    writer = open(outFile, "w")
    for row in lines:
        if len(row)>1:
            if "#" not in row[:1]:
                rowBits = row.split("\t")
                if way == "to_decimal":
                    RADeg = astCoords.hms2decimal(rowBits[RACol], delimiter)
                    decDeg = astCoords.dms2decimal(rowBits[decCol], delimiter)
                if way == "to_hmsdms":
                    RADeg = astCoords.decimal2hms(float(rowBits[RACol]),
                            delimiter)
                    decDeg = astCoords.decimal2dms(float(rowBits[decCol]),
                            delimiter)
                writeString = ""
                for i in range(len(rowBits)):
                    if i == RACol:
                        writeString = writeString+str(RADeg)+"\t"
                    elif i == decCol:
                        writeString = writeString+str(decDeg)+"\t"
                    elif rowBits[i].find("\n") != -1:
                        writeString = writeString+str(rowBits[i])
                    else:
                        writeString = writeString+str(rowBits[i])+"\t"
                # new line character already included
                writer.write(writeString.replace("\n", "").replace("\t",
                    colDelim)+"\n")
            else:
                writer.write(row.replace("\t", colDelim))
    writer.close()

#-----------------------------------------------------------------------------



#!/usr/bin/python
#File: NEDQueryPlot/NEDQueryPlot.py
#Created: Sat Dec 15 17:16:16 2012
#Last Change: Sat Dec 15 17:23:11 2012
# -*- coding: utf-8 -*-
#
# Fetches DSS image at a given position from Skyview, makes a plot listing NED
# matches found in image

#-----------------------------------------------------------------------------
import sys
import os
import math
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve
from astLib import *
import astropy.io.fits as pyfits
import pylab

#-----------------------------------------------------------------------------
def queryNED(RAMin, RAMax, decMin, decMax):
    """Queries NED, returns list of objects. RAMin, RAMax, decMin, decMax, all
    in degrees

    """

    # Just galaxies, groups
    urlretrieve("http://nedwww.ipac.caltech.edu/cgi-bin/nph-allsky?ra_constraint=Between&ra_1=%.6fd&ra_2=%.6fd&dec_constraint=Between&dec_1=%.6fd&dec_2=%.6fd&glon_constraint=Unconstrained&glon_1=&glon_2=&glat_constraint=Unconstrained&glat_1=&glat_2=&hconst=73&omegam=0.27&omegav=0.73&corr_z=1&z_constraint=Unconstrained&z_value1=&z_value2=&z_unit=z&flux_constraint=Unconstrained&flux_value1=&flux_value2=&flux_unit=Jy&frat_constraint=Unconstrained&ot_include=ANY&in_objtypes1=GGroups&in_objtypes1=Galaxies&nmp_op=ANY&out_csys=Equatorial&out_equinox=J2000.0&obj_sort=RA+or+Longitude&of=ascii_tab&zv_breaker=30000.0&list_limit=5&img_stamp=YES"
    % (RAMin, RAMax, decMin, decMax), "result.txt")

    inFile = open("result.txt", "r")
    lines = inFile.readlines()

    dataStarted = False
    labels = []
    names = []
    RAs = []
    decs = []
    sourceTypes = []
    redshifts = []
    for line in lines:
        bits = line.split("\t")
        if bits[0] == "1":
            dataStarted = True
        if dataStarted == True:
            labels.append(bits[0])
            names.append(bits[1])
            RAs.append(float(bits[2]))
            decs.append(float(bits[3]))
            sourceTypes.append(str(bits[4]))
            if bits[6] == '':
                redshifts.append('N/A')
            else:
                redshifts.append(str(bits[6]))

    return ({'labels': labels, 'names': names, 'RAs': RAs, 'decs': decs,
            'sourceTypes': sourceTypes, 'redshifts': redshifts})
    #-------------------------------------------------------------------------
# Main

# Query NED at this position - it's Stephan's quintet
RADeg = 338.98963
decDeg = 33.95991
name = "Stephan's Quintet"

if os.path.exists("fitsImages") == False:
    os.makedirs("fitsImages")
fitsFolder = "fitsImages"

outImageName = name.replace(" ", "_")+".png"
outFITS = name.replace(" ", "_")+".fits"

print("... fetching image from skyview ...")
if os.path.exists(fitsFolder+os.path.sep+outFITS) == True:
    os.remove(fitsFolder+os.path.sep+outFITS)
urlretrieve("http://skyview.gsfc.nasa.gov/cgi-bin/images?Position="+str(RADeg)+","+str(decDeg)+"&Size=0.1&Pixels=1000&Projection=Tan&Grid=J2000&Survey=digitized+sky+survey&Coordinates=J2000&Return=fits",
fitsFolder+os.path.sep+outFITS)

img = pyfits.open(fitsFolder+os.path.sep+outFITS)
wcs = astWCS.WCS(fitsFolder+os.path.sep+outFITS)
imgData = img[0].data
RAMin, RAMax, decMin, decMax=wcs.getImageMinMaxWCSCoords()
objs = queryNED(RAMin, RAMax, decMin, decMax)

fig = pylab.figure(figsize=(10,14))
pylab.figtext(0.5, 0.96, name, fontsize=20, horizontalalignment='center')
pylab.figtext(0.5, 0.94, "R.A, Dec. (J2000) = %s, %s" %
    (astCoords.decimal2hms(RADeg, ":"), astCoords.decimal2dms(decDeg, ":")),
    fontsize=14, horizontalalignment='center')

plot = astPlots.ImagePlot(imgData, wcs, cutLevels=[2000,7000], axes=[0.125,
        0.425, 0.8, 0.5])
plot.addPlotObjects([RADeg], [decDeg], 'actPosition', symbol='cross',
    color='cyan', size=8.0)

pylab.figtext(0.125, 0.35, "NED Matches:", fontweight='bold', fontsize=16)
pylab.figtext(0.125, 0.325, "ID", fontsize=14, horizontalalignment='left',
    verticalalignment='baseline', fontweight='bold')
pylab.figtext(0.16, 0.325, "Name", fontsize=14, horizontalalignment='left',
    verticalalignment='baseline', fontweight='bold')
pylab.figtext(0.42, 0.325, "R.A. (deg.)", fontsize=14,
    horizontalalignment='left', verticalalignment='baseline',
    fontweight='bold')
pylab.figtext(0.57, 0.325, "Dec. (deg.)", fontsize=14,
    horizontalalignment='left', verticalalignment='baseline',
    fontweight='bold')
pylab.figtext(0.72, 0.325, "Type", fontsize=14, horizontalalignment='left',
    verticalalignment='baseline', fontweight='bold')
pylab.figtext(0.8, 0.325, "Redshift", fontsize=14, horizontalalignment='left',
    verticalalignment='baseline', fontweight='bold')
rowSpacing=0.02

if len(objs['RAs']) > 0:
    plot.addPlotObjects(objs['RAs'], objs['decs'], 'nedObjects',
        objLabels=objs['labels'], size=8.0)
    for i in range(len(objs['RAs'])):
        pylab.figtext(0.125, 0.32-(i+1)*rowSpacing, objs['labels'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
        pylab.figtext(0.16, 0.32-(i+1)*rowSpacing, objs['names'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
        pylab.figtext(0.42, 0.32-(i+1)*rowSpacing, "%.6f" % objs['RAs'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
        pylab.figtext(0.57, 0.32-(i+1)*rowSpacing, "%.6f" % objs['decs'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
        pylab.figtext(0.72, 0.32-(i+1)*rowSpacing, objs['sourceTypes'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
        pylab.figtext(0.8, 0.32-(i+1)*rowSpacing, objs['redshifts'][i],
            fontsize=10, horizontalalignment='left',
            verticalalignment='baseline')
else:
    pylab.figtext(0.5, 0.32-rowSpacing, 'No matches found', fontsize=10,
        fontstyle='italic', horizontalalignment='center',
        verticalalignment='baseline')

plot.draw()
plot.save(outImageName)

pylab.close()
img.close()



"""Module for performing statistical calculations.

(c) 2007-2012 Matt Hilton 

(c) 2013-2014 Matt Hilton & Steven Boada

This module (as you may notice) provides very few statistical routines. It does, however, provide
biweight (robust) estimators of location and scale, as described in Beers et al. 1990 (AJ, 100,
32), in addition to a robust least squares fitting routine that uses the biweight transform.

Some routines may fail if they are passed lists with few items and encounter a `divide by zero'
error. Where this occurs, the function will return None. An error message will be printed to the
console when this happens if astStats.REPORT_ERRORS=True (the default). Testing if an
astStats function returns None can be used to handle errors in scripts. 

For extensive statistics modules, the Python bindings for GNU R (U{http://rpy.sourceforge.net}), or
SciPy (U{http://www.scipy.org}) are suggested.

"""

import math
import numpy
import sys

REPORT_ERRORS=True

#---------------------------------------------------------------------------------------------------
def mean(dataList):
    """Calculates the mean average of a list of numbers.
    
    @type dataList: list or numpy array
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: mean average
    
    """
    return numpy.mean(dataList)
    
#---------------------------------------------------------------------------------------------------
def weightedMean(dataList):
    """Calculates the weighted mean average of a two dimensional list (value, weight) of
    numbers.
    
    @type dataList: list
    @param dataList: input data, must be a two dimensional list in format [value, weight]
    @rtype: float
    @return: weighted mean average
    
    """
    sum=0
    weightSum=0
    for item in dataList:
        sum=sum+float(item[0]*item[1])
        weightSum=weightSum+item[1]
    if len(dataList)>0:
        mean=sum/weightSum
    else:
        mean=0
    return mean

#---------------------------------------------------------------------------------------------------
def stdev(dataList):
    """Calculates the (sample) standard deviation of a list of numbers.
    
    @type dataList: list or numpy array
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: standard deviation
    
    """
    return numpy.std(dataList)
    
#---------------------------------------------------------------------------------------------------
def rms(dataList):
    """Calculates the root mean square of a list of numbers.
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: root mean square
    
    """
    dataListSq=[]
    for item in dataList:
        dataListSq.append(item*item)
    listMeanSq=mean(dataListSq)
    rms=math.sqrt(listMeanSq)

    return rms
        
#---------------------------------------------------------------------------------------------------
def weightedStdev(dataList):
    """Calculates the weighted (sample) standard deviation of a list of numbers. 
    
    @type dataList: list
    @param dataList: input data, must be a two dimensional list in format [value, weight]
    @rtype: float
    @return: weighted standard deviation
    
    @note: Returns None if an error occurs.
    
    """
    listMean=weightedMean(dataList)
    sum=0
    wSum=0
    wNonZero=0
    for item in dataList:
        if item[1]>0.0:
            sum=sum+float((item[0]-listMean)/item[1])*float((item[0]-listMean)/item[1])
            wSum=wSum+float(1.0/item[1])*float(1.0/item[1])
            
    if len(dataList)>1:
        nFactor=float(len(dataList))/float(len(dataList)-1)
        stdev=math.sqrt(nFactor*(sum/wSum))
    else:
        if REPORT_ERRORS==True:
            print("""ERROR: astStats.weightedStdev() : dataList contains < 2 items.""")
        stdev=None
    return stdev
        
#---------------------------------------------------------------------------------------------------
def median(dataList):
    """Calculates the median of a list of numbers.
    
    @type dataList: list or numpy array
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: median average
    
    """        
    return numpy.median(dataList)
    
#---------------------------------------------------------------------------------------------------
def modeEstimate(dataList):
    """Returns an estimate of the mode of a set of values by mode=(3*median)-(2*mean).
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: estimate of mode average
    
    """
    mode=(3*median(dataList))-(2*mean(dataList))

    return mode

#---------------------------------------------------------------------------------------------------
def MAD(dataList):
    """Calculates the Median Absolute Deviation of a list of numbers.
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @rtype: float
    @return: median absolute deviation
    
    """
    listMedian=median(dataList)
    
    # Calculate |x-M| values
    diffModuli=[]
    for item in dataList:
        diffModuli.append(math.fabs(item-listMedian))
    
    MAD=median(diffModuli)
        
    return MAD

#---------------------------------------------------------------------------------------------------
def biweightLocation(dataList, tuningConstant):
    """Calculates the biweight location estimator (like a robust average) of a list of
    numbers.
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @type tuningConstant: float
    @param tuningConstant: 6.0 is recommended.
    @rtype: float
    @return: biweight location
    
    @note: Returns None if an error occurs.	
    
    """	
    C=tuningConstant
    listMedian=median(dataList)
    listMAD=MAD(dataList)
    if listMAD!=0:
        uValues=[]
        for item in dataList:
            uValues.append((item-listMedian)/(C*listMAD))
                
        top=0		# numerator equation (5) Beers et al if you like
        bottom=0	# denominator
        for i in range(len(uValues)):
            if math.fabs(uValues[i])<=1.0:
                top=top+((dataList[i]-listMedian) \
                    *(1.0-(uValues[i]*uValues[i])) \
                    *(1.0-(uValues[i]*uValues[i])))
            
                bottom=bottom+((1.0-(uValues[i]*uValues[i])) \
                    *(1.0-(uValues[i]*uValues[i])))
    
        CBI=listMedian+(top/bottom)
        
    else:
        if REPORT_ERRORS==True:
            print("""ERROR: astStats: biweightLocation() : MAD() returned 0.""")
        return None
    
    return CBI

#---------------------------------------------------------------------------------------------------
def biweightScale(dataList, tuningConstant):
    """Calculates the biweight scale estimator (like a robust standard deviation) of a list
    of numbers. 
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @type tuningConstant: float
    @param tuningConstant: 9.0 is recommended.
    @rtype: float
    @return: biweight scale
    
    @note: Returns None if an error occurs.
        
    """	
    C=tuningConstant
    
    # Calculate |x-M| values and u values
    listMedian=median(dataList)
    listMAD=MAD(dataList)
    diffModuli=[]
    for item in dataList:
        diffModuli.append(math.fabs(item-listMedian))
    uValues=[]
    for item in dataList:
        try:
            uValues.append((item-listMedian)/(C*listMAD))
        except ZeroDivisionError:
            if REPORT_ERRORS==True:
                print("""ERROR: astStats.biweightScale() : divide by zero error.""")
            return None
        
    top=0		# numerator equation (9) Beers et al
    bottom=0
    valCount=0	# Count values where u<1 only
    
    for i in range(len(uValues)):
        # Skip u values >1
        if math.fabs(uValues[i])<=1.0:
            u2Term=1.0-(uValues[i]*uValues[i])
            u4Term=math.pow(u2Term, 4)
            top=top+((diffModuli[i]*diffModuli[i])*u4Term)
            bottom=bottom+(u2Term*(1.0-(5.0*(uValues[i]*uValues[i]))))
            valCount=valCount+1
    
    top=math.sqrt(top)
    bottom=math.fabs(bottom)

    SBI=math.pow(float(valCount), 0.5)*(top/bottom)
    return SBI

#---------------------------------------------------------------------------------------------------
def biweightClipped(dataList, tuningConstant, sigmaCut):
    """Iteratively calculates biweight location and scale, using sigma clipping, for a list
    of values. 	The calculation is performed on the first column of a multi-dimensional
    list; other columns are ignored.
    
    @type dataList: list
    @param dataList: input data, must contain more than 5 values
    @type tuningConstant: float
    @param tuningConstant: 6.0 is recommended for location estimates, 9.0 is recommended for
    scale estimates	
    @type sigmaCut: float
    @param sigmaCut: sigma clipping to apply
    @rtype:	dictionary 
    @return: estimate of biweight location, scale, and list of non-clipped data, in the format
    {'biweightLocation', 'biweightScale', 'dataList'}
    
    @note: Returns None if an error occurs.

    """		
    
    iterations=0
    clippedValues=[]
    for row in dataList:
        if type(row)==list:
            clippedValues.append(row[0])
        else:
            clippedValues.append(row)
    
    cbi=None
    sbi=None
    clippedData=None
    while iterations<11 and len(clippedValues)>5:
        
        cbi=biweightLocation(clippedValues, tuningConstant)	
        sbi=biweightScale(clippedValues, tuningConstant)
        
        # check for either biweight routine falling over
        # happens when feed in lots of similar numbers
        # e.g. when bootstrapping with a small sample		
        if cbi==None or sbi==None:
            
            if REPORT_ERRORS==True:
                print("""ERROR: astStats : biweightClipped() :
                divide by zero error.""")
            
            return None
            
        else:
            
            clippedValues=[]
            clippedData=[]
            for row in dataList:
                if type(row)==list:
                    if row[0]>cbi-(sigmaCut*sbi) \
                    and row[0]<cbi+(sigmaCut*sbi):
                        clippedValues.append(row[0])
                        clippedData.append(row)
                else:
                    if row>cbi-(sigmaCut*sbi) \
                    and row<cbi+(sigmaCut*sbi):
                        clippedValues.append(row)
                        clippedData.append(row)
            
        iterations=iterations+1
            
    return {'biweightLocation':cbi, 'biweightScale':sbi, 'dataList':clippedData}

#---------------------------------------------------------------------------------------------------
def biweightTransform(dataList, tuningConstant):
    """Calculates the biweight transform for a set of values. Useful for using as weights in
    robust line fitting.
    
    @type dataList: list
    @param dataList: input data, must be a one dimensional list
    @type tuningConstant: float
    @param tuningConstant: 6.0 is recommended for location estimates, 9.0 is recommended for
    scale estimates	
    @rtype: list
    @return: list of biweights	
    
    """
    C=tuningConstant
    
    # Calculate |x-M| values and u values
    listMedian=abs(median(dataList))
    cutoff=C*listMedian
    biweights=[]
    for item in dataList:
        if abs(item)<cutoff:
            biweights.append([item,
            (1.0-((item/cutoff)*(item/cutoff))) \
            *(1.0-((item/cutoff)*(item/cutoff)))])
        else:
            biweights.append([item, 0.0])
    
    return biweights
    
#---------------------------------------------------------------------------------------------------
def OLSFit(dataList):
    """Performs an ordinary least squares fit on a two dimensional list of numbers.
    Minimum number of data points is 5.
    
    @type dataList: list
    @param dataList: input data, must be a two dimensional list in format [x, y]
    @rtype: dictionary
    @return: slope and intercept on y-axis, with associated errors, in the format
    {'slope', 'intercept', 'slopeError', 'interceptError'}
    
    @note: Returns None if an error occurs.	
        
    """
    sumX=0
    sumY=0
    sumXY=0
    sumXX=0
    n=float(len(dataList))
    if n > 2:
        for item in dataList:
            sumX=sumX+item[0]
            sumY=sumY+item[1]
            sumXY=sumXY+(item[0]*item[1])
            sumXX=sumXX+(item[0]*item[0])	
        m=((n*sumXY)-(sumX*sumY))/((n*sumXX)-(sumX*sumX))
        c=((sumXX*sumY)-(sumX*sumXY))/((n*sumXX)-(sumX*sumX))
        
        sumRes=0
        for item in dataList:
        
            sumRes=sumRes+((item[1]-(m*item[0])-c) \
            *(item[1]-(m*item[0])-c))
            
        sigma=math.sqrt((1.0/(n-2))*sumRes)
        
        try:
            mSigma=(sigma*math.sqrt(n))/math.sqrt((n*sumXX)-(sumX*sumX))
        except:
            mSigma=numpy.nan
        try:
            cSigma=(sigma*math.sqrt(sumXX))/math.sqrt((n*sumXX)-(sumX*sumX))
        except:
            cSigma=numpy.nan
    else:
        if REPORT_ERRORS==True:
            print("""ERROR: astStats.OLSFit() : dataList contains < 3 items.""")
            
        return None
        
    return {'slope':m,
            'intercept':c,
            'slopeError':mSigma,
            'interceptError':cSigma}

#---------------------------------------------------------------------------------------------------
def clippedMeanStdev(dataList, sigmaCut = 3.0, maxIterations = 10.0):
    """Calculates the clipped mean and stdev of a list of numbers.
    
    @type dataList: list
    @param dataList: input data, one dimensional list of numbers
    @type sigmaCut: float
    @param sigmaCut: clipping in Gaussian sigma to apply
    @type maxIterations: int
    @param maxIterations: maximum number of iterations
    @rtype: dictionary
    @return: format {'clippedMean', 'clippedStdev', 'numPoints'}
    
    """
    
    listCopy=[]
    for d in dataList:
        listCopy.append(d)
    listCopy=numpy.array(listCopy)
    
    iterations=0
    while iterations < maxIterations and len(listCopy) > 4:
        
        m=listCopy.mean()
        s=listCopy.std()
        
        listCopy=listCopy[numpy.less(abs(listCopy), abs(m+sigmaCut*s))]
        
        iterations=iterations+1
    
    return {'clippedMean': m, 'clippedStdev': s, 'numPoints': listCopy.shape[0]}
    
#---------------------------------------------------------------------------------------------------
def clippedWeightedLSFit(dataList, sigmaCut):
    """Performs a weighted least squares fit on a list of numbers with sigma clipping. Minimum number of data
    points is 5.
    
    @type dataList: list
    @param dataList: input data, must be a three dimensional list in format [x, y, y weight]
    @rtype: dictionary
    @return: slope and intercept on y-axis, with associated errors, in the format
    {'slope', 'intercept', 'slopeError', 'interceptError'}
    
    @note: Returns None if an error occurs.	
    
    """
    
    iterations=0
    clippedValues=[]
    for row in dataList:
        clippedValues.append(row)
        
    while iterations<11 and len(clippedValues)>4:
        
        fitResults=weightedLSFit(clippedValues, "errors")
        
        if fitResults['slope'] == None:
            
            if REPORT_ERRORS==True:
                print("""ERROR: astStats : clippedWeightedLSFit() :
                divide by zero error.""")
            
            return None
            
        else:
            
            clippedValues=[]
            for row in dataList:
                
                # Trim points more than sigmaCut*sigma away from the fitted line
                fit=fitResults['slope']*row[0]+fitResults['intercept']
                res=row[1]-fit
                if abs(res)/row[2] < sigmaCut:
                    clippedValues.append(row)
            
        iterations=iterations+1
    
    # store the number of values that made it through the clipping process
    fitResults['numDataPoints']=len(clippedValues)
    
    return fitResults
    
#---------------------------------------------------------------------------------------------------
def weightedLSFit(dataList, weightType):
    """Performs a weighted least squares fit on a three dimensional list of numbers [x, y, y error].
    
    @type dataList: list
    @param dataList: input data, must be a three dimensional list in format [x, y, y error]
    @type weightType: string
    @param weightType: if "errors", weights are calculated assuming the input data is in the
    format [x, y, error on y]; if "weights", the weights are assumed to be already calculated and
    stored in a fourth column [x, y, error on y, weight] (as used by e.g. L{astStats.biweightLSFit})
    @rtype: dictionary
    @return: slope and intercept on y-axis, with associated errors, in the format
    {'slope', 'intercept', 'slopeError', 'interceptError'}
    
    @note: Returns None if an error occurs.	
            
    """
    if weightType == "weights":
        sumW=0
        sumWX=0
        sumWY=0
        sumWXY=0
        sumWXX=0
        n=float(len(dataList))
        if n > 4:
            for item in dataList:
                W=item[3]
                sumWX=sumWX+(W*item[0])
                sumWY=sumWY+(W*item[1])
                sumWXY=sumWXY+(W*item[0]*item[1])
                sumWXX=sumWXX+(W*item[0]*item[0])
                sumW=sumW+W
                #print sumW, sumWXX, sumWX
        
            try:
                m=((sumW*sumWXY)-(sumWX*sumWY)) \
                /((sumW*sumWXX)-(sumWX*sumWX))
            except ZeroDivisionError:
                if REPORT_ERRORS == True:
                    print("ERROR: astStats.weightedLSFit() : divide by zero error.")
                return None
        
            try:
                c=((sumWXX*sumWY)-(sumWX*sumWXY)) \
                /((sumW*sumWXX)-(sumWX*sumWX))
            except ZeroDivisionError:
                if REPORT_ERRORS == True:
                    print("ERROR: astStats.weightedLSFit() : divide by zero error.")
                return None
            
            sumRes=0
            for item in dataList:
            
                sumRes=sumRes+((item[1]-(m*item[0])-c) \
                *(item[1]-(m*item[0])-c))
                
            sigma=math.sqrt((1.0/(n-2))*sumRes)
            
            # Can get div0 errors here so check
            # When biweight fitting converges this shouldn't happen
            if (n*sumWXX)-(sumWX*sumWX)>0.0: 
            
                mSigma=(sigma*math.sqrt(n)) \
                    /math.sqrt((n*sumWXX)-(sumWX*sumWX))
        
                cSigma=(sigma*math.sqrt(sumWXX)) \
                    /math.sqrt((n*sumWXX)-(sumWX*sumWX))
                
            else:
                
                if REPORT_ERRORS==True:
                    print("""ERROR: astStats.weightedLSFit()
                    : divide by zero error.""")
                return None
                
        else:
            if REPORT_ERRORS==True:
                print("""ERROR: astStats.weightedLSFit() :
                dataList contains < 5 items.""")
            return None
            
    elif weightType == "errors":
        sumX=0
        sumY=0
        sumXY=0
        sumXX=0
        sumSigma=0
        n=float(len(dataList))
        for item in dataList:
            sumX=sumX+(item[0]/(item[2]*item[2]))
            sumY=sumY+(item[1]/(item[2]*item[2]))
            sumXY=sumXY+((item[0]*item[1])/(item[2]*item[2]))
            sumXX=sumXX+((item[0]*item[0])/(item[2]*item[2]))
            sumSigma=sumSigma+(1.0/(item[2]*item[2]))
        delta=(sumSigma*sumXX)-(sumX*sumX)	
        m=((sumSigma*sumXY)-(sumX*sumY))/delta
        c=((sumXX*sumY)-(sumX*sumXY))/delta
        mSigma=math.sqrt(sumSigma/delta)
        cSigma=math.sqrt(sumXX/delta)
        
    return {'slope':m,
            'intercept':c,
            'slopeError':mSigma,
            'interceptError':cSigma}
    
#---------------------------------------------------------------------------------------------------
def biweightLSFit(dataList, tuningConstant, sigmaCut = None):
    """Performs a weighted least squares fit, where the weights used are the biweight
    transforms of the residuals to the previous best fit .i.e. the procedure is iterative,
    and converges very quickly (iterations is set to 10 by default). Minimum number of data
    points is 10.
    
    This seems to give slightly different results to the equivalent R routine, so use at your
    own risk!
    
    @type dataList: list
    @param dataList: input data, must be a three dimensional list in format [x, y, y weight]
    @type tuningConstant: float
    @param tuningConstant: 6.0 is recommended for location estimates, 9.0 is recommended for
    scale estimates
    @type sigmaCut: float
    @param sigmaCut: sigma clipping to apply (set to None if not required)	
    @rtype: dictionary
    @return: slope and intercept on y-axis, with associated errors, in the format
    {'slope', 'intercept', 'slopeError', 'interceptError'}
    
    @note: Returns None if an error occurs.
        
    """

    dataCopy=[]
    for row in dataList:
        dataCopy.append(row)
        
    # First perform unweighted fit, then calculate residuals
    results=OLSFit(dataCopy)
    origLen=len(dataCopy)
    for k in range(10):
        m=results['slope']
        c=results['intercept']
        res=[]
        for item in dataCopy:
            res.append((m*item[0]+c)-item[1])
            
        if len(res)>5:
            # For clipping, trim away things >3 sigma 
            # away from median
            if sigmaCut != None:
                absRes=[]
                for item in res:
                    absRes.append(abs(item))
                sigma=stdev(absRes)
                count=0
                for item in absRes:
                    if item>(sigmaCut*sigma) \
                    and len(dataCopy)>2:
                        del dataCopy[count]
                        del res[count]
                        
                        # Index of datalist gets out of
                        # sync with absRes as we delete
                        # items
                        count=count-1 
                        
                    count=count+1
                        
            # Biweight transform residuals
            weights=biweightTransform(res, tuningConstant)
                        
            # Perform weighted fit, using biweight transforms 
            # of residuals as weight
            wData=[]
            for i in range(len(dataCopy)):
                wData.append([dataCopy[i][0], dataCopy[i][1], dataCopy[i][2], weights[i][1]])
            
            results=weightedLSFit(wData, "weights")

    return results
    
#---------------------------------------------------------------------------------------------------
def cumulativeBinner(data, binMin, binMax, binTotal):
    """Bins the input data cumulatively.
    
    @param data: input data, must be a one dimensional list
    @type binMin: float
    @param binMin: minimum value from which to bin data
    @type binMax: float
    @param binMax: maximum value from which to bin data	
    @type binTotal: int
    @param binTotal: number of bins 
    @rtype: list
    @return: binned data, in format [bin centre, frequency]
        
    """
    #Bin data
    binStep=float(binMax-binMin)/binTotal
    bins=[]
    totalItems=len(data)
    for i in range(binTotal):
        bins.append(0)
        for item in data:
            if item>(binMin+(i*binStep)):
                bins[i]=bins[i]+1.0/totalItems
                
    # Gnuplot requires points at bin midpoints
    coords=[]
    for i in range(binTotal):
        coords.append([binMin+(float(i+0.5)*binStep), bins[i]])
    
    return coords

#---------------------------------------------------------------------------------------------------
def binner(data, binMin, binMax, binTotal):
    """Bins the input data..
    
    @param data: input data, must be a one dimensional list
    @type binMin: float
    @param binMin: minimum value from which to bin data
    @type binMax: float
    @param binMax: maximum value from which to bin data	
    @type binTotal: int
    @param binTotal: number of bins 
    @rtype: list
    @return: binned data, in format [bin centre, frequency]
        
    """
    #Bin data
    binStep=float(binMax-binMin)/binTotal
    bins=[]
    for i in range(binTotal):
        bins.append(0)
        for item in data:
            if item>(binMin+(i*binStep)) \
            and item<=(binMin+((i+1)*binStep)):
                bins[i]=bins[i]+1
                
    # Gnuplot requires points at bin midpoints
    coords=[]
    for i in range(binTotal):
        coords.append([binMin+(float(i+0.5)*binStep), bins[i]])
    
    return coords

#---------------------------------------------------------------------------------------------------
def weightedBinner(data, weights, binMin, binMax, binTotal):
    """Bins the input data, recorded frequency is sum of weights in bin.
    
    @param data: input data, must be a one dimensional list
    @type binMin: float
    @param binMin: minimum value from which to bin data
    @type binMax: float
    @param binMax: maximum value from which to bin data	
    @type binTotal: int
    @param binTotal: number of bins 
    @rtype: list
    @return: binned data, in format [bin centre, frequency]
        
    """
    #Bin data
    binStep=float(binMax-binMin)/binTotal
    bins=[]
    for i in range(binTotal):
        bins.append(0.0)
        for item, weight in zip(data, weights):
            if item>(binMin+(i*binStep)) \
            and item<=(binMin+((i+1)*binStep)):
                bins[i]=bins[i]+weight
                
    # Gnuplot requires points at bin midpoints
    coords=[]
    for i in range(binTotal):
        coords.append([binMin+(float(i+0.5)*binStep), bins[i]])
    
    return coords
    
#---------------------------------------------------------------------------------------------------



#!/usr/bin/env python
#File: rotatedClipTest/rotatedClipTest.py
#Created: Sat Dec 15 17:05:45 2012
#Last Change: Sat Dec 15 17:06:28 2012
# -*- coding: utf-8 -*-
#
# Quick script to test out rotation, clipping

import astropy.io.fits as pyfits
from astLib import *

img = pyfits.open("../../../testingData/testImage1.fits")
wcs = astWCS.WCS("../../../testingData/testImage1.fits")
d = img[0].data

# Clip slightly off centre
RADeg, decDeg = wcs.getCentreWCSCoords()
RADeg = RADeg+1.2/60.0
decDeg = decDeg-1.0/60.0

clip = astImages.clipRotatedImageSectionWCS(d, wcs, RADeg, decDeg, 3.0/60.0)
clipnr = astImages.clipImageSectionWCS(d, wcs, RADeg, decDeg, 3.0/60.0)

astImages.saveFITS("output_rotated.fits", clip['data'], clip['wcs'])
astImages.saveFITS("output_notRotated.fits", clipnr['data'], clipnr['wcs'])



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.2.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _wcs
else:
    import _wcs

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)



def new_doubleArray(nelements):
    r"""new_doubleArray(nelements) -> double *"""
    return _wcs.new_doubleArray(nelements)

def delete_doubleArray(ary):
    r"""delete_doubleArray(ary)"""
    return _wcs.delete_doubleArray(ary)

def doubleArray_getitem(ary, index):
    r"""doubleArray_getitem(ary, index) -> double"""
    return _wcs.doubleArray_getitem(ary, index)

def doubleArray_setitem(ary, index, value):
    r"""doubleArray_setitem(ary, index, value)"""
    return _wcs.doubleArray_setitem(ary, index, value)

def wcsinit(hstring):
    r"""wcsinit(hstring) -> WorldCoor"""
    return _wcs.wcsinit(hstring)

def wcsxinit(cra, cdec, secpix, xrpix, yrpix, nxpix, nypix, rotate, equinox, epoch, proj):
    r"""wcsxinit(cra, cdec, secpix, xrpix, yrpix, nxpix, nypix, rotate, equinox, epoch, proj) -> WorldCoor"""
    return _wcs.wcsxinit(cra, cdec, secpix, xrpix, yrpix, nxpix, nypix, rotate, equinox, epoch, proj)

def wcskinit(nxpix, nypix, ctype1, ctype2, crpix1, crpix2, crval1, crval2, cd, cdelt1, cdelt2, crota, equinox, epoch):
    r"""wcskinit(nxpix, nypix, ctype1, ctype2, crpix1, crpix2, crval1, crval2, cd, cdelt1, cdelt2, crota, equinox, epoch) -> WorldCoor"""
    return _wcs.wcskinit(nxpix, nypix, ctype1, ctype2, crpix1, crpix2, crval1, crval2, cd, cdelt1, cdelt2, crota, equinox, epoch)

def iswcs(wcs):
    r"""iswcs(wcs) -> int"""
    return _wcs.iswcs(wcs)

def nowcs(wcs):
    r"""nowcs(wcs) -> int"""
    return _wcs.nowcs(wcs)

def wcs2pix(wcs, xpos, ypos):
    r"""wcs2pix(wcs, xpos, ypos)"""
    return _wcs.wcs2pix(wcs, xpos, ypos)

def pix2wcs(wcs, xpix, ypix):
    r"""pix2wcs(wcs, xpix, ypix)"""
    return _wcs.pix2wcs(wcs, xpix, ypix)

def wcscent(wcs):
    r"""wcscent(wcs)"""
    return _wcs.wcscent(wcs)

def getradecsys(wcs):
    r"""getradecsys(wcs) -> char *"""
    return _wcs.getradecsys(wcs)

def wcsoutinit(wcs, coorsys):
    r"""wcsoutinit(wcs, coorsys)"""
    return _wcs.wcsoutinit(wcs, coorsys)

def wcsininit(wcs, coorsys):
    r"""wcsininit(wcs, coorsys)"""
    return _wcs.wcsininit(wcs, coorsys)

def getwcsout(wcs):
    r"""getwcsout(wcs) -> char *"""
    return _wcs.getwcsout(wcs)

def getwcsin(wcs):
    r"""getwcsin(wcs) -> char *"""
    return _wcs.getwcsin(wcs)

def wcssize(wcs):
    r"""wcssize(wcs)"""
    return _wcs.wcssize(wcs)

def wcsfull(wcs):
    r"""wcsfull(wcs)"""
    return _wcs.wcsfull(wcs)
class WorldCoor(object):
    r"""Proxy of C WorldCoor struct."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    xref = property(_wcs.WorldCoor_xref_get, _wcs.WorldCoor_xref_set, doc=r"""xref""")
    yref = property(_wcs.WorldCoor_yref_get, _wcs.WorldCoor_yref_set, doc=r"""yref""")
    xrefpix = property(_wcs.WorldCoor_xrefpix_get, _wcs.WorldCoor_xrefpix_set, doc=r"""xrefpix""")
    yrefpix = property(_wcs.WorldCoor_yrefpix_get, _wcs.WorldCoor_yrefpix_set, doc=r"""yrefpix""")
    xinc = property(_wcs.WorldCoor_xinc_get, _wcs.WorldCoor_xinc_set, doc=r"""xinc""")
    yinc = property(_wcs.WorldCoor_yinc_get, _wcs.WorldCoor_yinc_set, doc=r"""yinc""")
    rot = property(_wcs.WorldCoor_rot_get, _wcs.WorldCoor_rot_set, doc=r"""rot""")
    cd = property(_wcs.WorldCoor_cd_get, _wcs.WorldCoor_cd_set, doc=r"""cd""")
    dc = property(_wcs.WorldCoor_dc_get, _wcs.WorldCoor_dc_set, doc=r"""dc""")
    equinox = property(_wcs.WorldCoor_equinox_get, _wcs.WorldCoor_equinox_set, doc=r"""equinox""")
    epoch = property(_wcs.WorldCoor_epoch_get, _wcs.WorldCoor_epoch_set, doc=r"""epoch""")
    nxpix = property(_wcs.WorldCoor_nxpix_get, _wcs.WorldCoor_nxpix_set, doc=r"""nxpix""")
    nypix = property(_wcs.WorldCoor_nypix_get, _wcs.WorldCoor_nypix_set, doc=r"""nypix""")
    plate_ra = property(_wcs.WorldCoor_plate_ra_get, _wcs.WorldCoor_plate_ra_set, doc=r"""plate_ra""")
    plate_dec = property(_wcs.WorldCoor_plate_dec_get, _wcs.WorldCoor_plate_dec_set, doc=r"""plate_dec""")
    plate_scale = property(_wcs.WorldCoor_plate_scale_get, _wcs.WorldCoor_plate_scale_set, doc=r"""plate_scale""")
    x_pixel_offset = property(_wcs.WorldCoor_x_pixel_offset_get, _wcs.WorldCoor_x_pixel_offset_set, doc=r"""x_pixel_offset""")
    y_pixel_offset = property(_wcs.WorldCoor_y_pixel_offset_get, _wcs.WorldCoor_y_pixel_offset_set, doc=r"""y_pixel_offset""")
    x_pixel_size = property(_wcs.WorldCoor_x_pixel_size_get, _wcs.WorldCoor_x_pixel_size_set, doc=r"""x_pixel_size""")
    y_pixel_size = property(_wcs.WorldCoor_y_pixel_size_get, _wcs.WorldCoor_y_pixel_size_set, doc=r"""y_pixel_size""")
    ppo_coeff = property(_wcs.WorldCoor_ppo_coeff_get, _wcs.WorldCoor_ppo_coeff_set, doc=r"""ppo_coeff""")
    x_coeff = property(_wcs.WorldCoor_x_coeff_get, _wcs.WorldCoor_x_coeff_set, doc=r"""x_coeff""")
    y_coeff = property(_wcs.WorldCoor_y_coeff_get, _wcs.WorldCoor_y_coeff_set, doc=r"""y_coeff""")
    xpix = property(_wcs.WorldCoor_xpix_get, _wcs.WorldCoor_xpix_set, doc=r"""xpix""")
    ypix = property(_wcs.WorldCoor_ypix_get, _wcs.WorldCoor_ypix_set, doc=r"""ypix""")
    zpix = property(_wcs.WorldCoor_zpix_get, _wcs.WorldCoor_zpix_set, doc=r"""zpix""")
    xpos = property(_wcs.WorldCoor_xpos_get, _wcs.WorldCoor_xpos_set, doc=r"""xpos""")
    ypos = property(_wcs.WorldCoor_ypos_get, _wcs.WorldCoor_ypos_set, doc=r"""ypos""")
    crpix = property(_wcs.WorldCoor_crpix_get, _wcs.WorldCoor_crpix_set, doc=r"""crpix""")
    crval = property(_wcs.WorldCoor_crval_get, _wcs.WorldCoor_crval_set, doc=r"""crval""")
    cdelt = property(_wcs.WorldCoor_cdelt_get, _wcs.WorldCoor_cdelt_set, doc=r"""cdelt""")
    pc = property(_wcs.WorldCoor_pc_get, _wcs.WorldCoor_pc_set, doc=r"""pc""")
    projp = property(_wcs.WorldCoor_projp_get, _wcs.WorldCoor_projp_set, doc=r"""projp""")
    pvfail = property(_wcs.WorldCoor_pvfail_get, _wcs.WorldCoor_pvfail_set, doc=r"""pvfail""")
    projppv = property(_wcs.WorldCoor_projppv_get, _wcs.WorldCoor_projppv_set, doc=r"""projppv""")
    inv_x = property(_wcs.WorldCoor_inv_x_get, _wcs.WorldCoor_inv_x_set, doc=r"""inv_x""")
    inv_y = property(_wcs.WorldCoor_inv_y_get, _wcs.WorldCoor_inv_y_set, doc=r"""inv_y""")
    longpole = property(_wcs.WorldCoor_longpole_get, _wcs.WorldCoor_longpole_set, doc=r"""longpole""")
    latpole = property(_wcs.WorldCoor_latpole_get, _wcs.WorldCoor_latpole_set, doc=r"""latpole""")
    rodeg = property(_wcs.WorldCoor_rodeg_get, _wcs.WorldCoor_rodeg_set, doc=r"""rodeg""")
    imrot = property(_wcs.WorldCoor_imrot_get, _wcs.WorldCoor_imrot_set, doc=r"""imrot""")
    pa_north = property(_wcs.WorldCoor_pa_north_get, _wcs.WorldCoor_pa_north_set, doc=r"""pa_north""")
    pa_east = property(_wcs.WorldCoor_pa_east_get, _wcs.WorldCoor_pa_east_set, doc=r"""pa_east""")
    radvel = property(_wcs.WorldCoor_radvel_get, _wcs.WorldCoor_radvel_set, doc=r"""radvel""")
    zvel = property(_wcs.WorldCoor_zvel_get, _wcs.WorldCoor_zvel_set, doc=r"""zvel""")
    zpzd = property(_wcs.WorldCoor_zpzd_get, _wcs.WorldCoor_zpzd_set, doc=r"""zpzd""")
    zpr = property(_wcs.WorldCoor_zpr_get, _wcs.WorldCoor_zpr_set, doc=r"""zpr""")
    imflip = property(_wcs.WorldCoor_imflip_get, _wcs.WorldCoor_imflip_set, doc=r"""imflip""")
    prjcode = property(_wcs.WorldCoor_prjcode_get, _wcs.WorldCoor_prjcode_set, doc=r"""prjcode""")
    latbase = property(_wcs.WorldCoor_latbase_get, _wcs.WorldCoor_latbase_set, doc=r"""latbase""")
    ncoeff1 = property(_wcs.WorldCoor_ncoeff1_get, _wcs.WorldCoor_ncoeff1_set, doc=r"""ncoeff1""")
    ncoeff2 = property(_wcs.WorldCoor_ncoeff2_get, _wcs.WorldCoor_ncoeff2_set, doc=r"""ncoeff2""")
    zpnp = property(_wcs.WorldCoor_zpnp_get, _wcs.WorldCoor_zpnp_set, doc=r"""zpnp""")
    changesys = property(_wcs.WorldCoor_changesys_get, _wcs.WorldCoor_changesys_set, doc=r"""changesys""")
    printsys = property(_wcs.WorldCoor_printsys_get, _wcs.WorldCoor_printsys_set, doc=r"""printsys""")
    ndec = property(_wcs.WorldCoor_ndec_get, _wcs.WorldCoor_ndec_set, doc=r"""ndec""")
    degout = property(_wcs.WorldCoor_degout_get, _wcs.WorldCoor_degout_set, doc=r"""degout""")
    tabsys = property(_wcs.WorldCoor_tabsys_get, _wcs.WorldCoor_tabsys_set, doc=r"""tabsys""")
    rotmat = property(_wcs.WorldCoor_rotmat_get, _wcs.WorldCoor_rotmat_set, doc=r"""rotmat""")
    coorflip = property(_wcs.WorldCoor_coorflip_get, _wcs.WorldCoor_coorflip_set, doc=r"""coorflip""")
    offscl = property(_wcs.WorldCoor_offscl_get, _wcs.WorldCoor_offscl_set, doc=r"""offscl""")
    wcson = property(_wcs.WorldCoor_wcson_get, _wcs.WorldCoor_wcson_set, doc=r"""wcson""")
    naxis = property(_wcs.WorldCoor_naxis_get, _wcs.WorldCoor_naxis_set, doc=r"""naxis""")
    naxes = property(_wcs.WorldCoor_naxes_get, _wcs.WorldCoor_naxes_set, doc=r"""naxes""")
    wcsproj = property(_wcs.WorldCoor_wcsproj_get, _wcs.WorldCoor_wcsproj_set, doc=r"""wcsproj""")
    linmode = property(_wcs.WorldCoor_linmode_get, _wcs.WorldCoor_linmode_set, doc=r"""linmode""")
    detector = property(_wcs.WorldCoor_detector_get, _wcs.WorldCoor_detector_set, doc=r"""detector""")
    instrument = property(_wcs.WorldCoor_instrument_get, _wcs.WorldCoor_instrument_set, doc=r"""instrument""")
    ctype = property(_wcs.WorldCoor_ctype_get, _wcs.WorldCoor_ctype_set, doc=r"""ctype""")
    c1type = property(_wcs.WorldCoor_c1type_get, _wcs.WorldCoor_c1type_set, doc=r"""c1type""")
    c2type = property(_wcs.WorldCoor_c2type_get, _wcs.WorldCoor_c2type_set, doc=r"""c2type""")
    ptype = property(_wcs.WorldCoor_ptype_get, _wcs.WorldCoor_ptype_set, doc=r"""ptype""")
    units = property(_wcs.WorldCoor_units_get, _wcs.WorldCoor_units_set, doc=r"""units""")
    radecsys = property(_wcs.WorldCoor_radecsys_get, _wcs.WorldCoor_radecsys_set, doc=r"""radecsys""")
    radecout = property(_wcs.WorldCoor_radecout_get, _wcs.WorldCoor_radecout_set, doc=r"""radecout""")
    radecin = property(_wcs.WorldCoor_radecin_get, _wcs.WorldCoor_radecin_set, doc=r"""radecin""")
    eqin = property(_wcs.WorldCoor_eqin_get, _wcs.WorldCoor_eqin_set, doc=r"""eqin""")
    eqout = property(_wcs.WorldCoor_eqout_get, _wcs.WorldCoor_eqout_set, doc=r"""eqout""")
    sysin = property(_wcs.WorldCoor_sysin_get, _wcs.WorldCoor_sysin_set, doc=r"""sysin""")
    syswcs = property(_wcs.WorldCoor_syswcs_get, _wcs.WorldCoor_syswcs_set, doc=r"""syswcs""")
    sysout = property(_wcs.WorldCoor_sysout_get, _wcs.WorldCoor_sysout_set, doc=r"""sysout""")
    center = property(_wcs.WorldCoor_center_get, _wcs.WorldCoor_center_set, doc=r"""center""")
    wcsl = property(_wcs.WorldCoor_wcsl_get, _wcs.WorldCoor_wcsl_set, doc=r"""wcsl""")
    lin = property(_wcs.WorldCoor_lin_get, _wcs.WorldCoor_lin_set, doc=r"""lin""")
    cel = property(_wcs.WorldCoor_cel_get, _wcs.WorldCoor_cel_set, doc=r"""cel""")
    prj = property(_wcs.WorldCoor_prj_get, _wcs.WorldCoor_prj_set, doc=r"""prj""")
    lngcor = property(_wcs.WorldCoor_lngcor_get, _wcs.WorldCoor_lngcor_set, doc=r"""lngcor""")
    latcor = property(_wcs.WorldCoor_latcor_get, _wcs.WorldCoor_latcor_set, doc=r"""latcor""")
    distcode = property(_wcs.WorldCoor_distcode_get, _wcs.WorldCoor_distcode_set, doc=r"""distcode""")
    distort = property(_wcs.WorldCoor_distort_get, _wcs.WorldCoor_distort_set, doc=r"""distort""")
    command_format = property(_wcs.WorldCoor_command_format_get, _wcs.WorldCoor_command_format_set, doc=r"""command_format""")
    ltm = property(_wcs.WorldCoor_ltm_get, _wcs.WorldCoor_ltm_set, doc=r"""ltm""")
    ltv = property(_wcs.WorldCoor_ltv_get, _wcs.WorldCoor_ltv_set, doc=r"""ltv""")
    idpix = property(_wcs.WorldCoor_idpix_get, _wcs.WorldCoor_idpix_set, doc=r"""idpix""")
    ndpix = property(_wcs.WorldCoor_ndpix_get, _wcs.WorldCoor_ndpix_set, doc=r"""ndpix""")
    wcs = property(_wcs.WorldCoor_wcs_get, _wcs.WorldCoor_wcs_set, doc=r"""wcs""")
    wcsdep = property(_wcs.WorldCoor_wcsdep_get, _wcs.WorldCoor_wcsdep_set, doc=r"""wcsdep""")
    wcsname = property(_wcs.WorldCoor_wcsname_get, _wcs.WorldCoor_wcsname_set, doc=r"""wcsname""")
    wcschar = property(_wcs.WorldCoor_wcschar_get, _wcs.WorldCoor_wcschar_set, doc=r"""wcschar""")
    logwcs = property(_wcs.WorldCoor_logwcs_get, _wcs.WorldCoor_logwcs_set, doc=r"""logwcs""")

    def __init__(self):
        r"""__init__(self) -> WorldCoor"""
        _wcs.WorldCoor_swiginit(self, _wcs.new_WorldCoor())
    __swig_destroy__ = _wcs.delete_WorldCoor

# Register WorldCoor in _wcs:
_wcs.WorldCoor_swigregister(WorldCoor)




"""Module for handling World Coordinate Systems (WCS).

(c) 2007-2012 Matt Hilton

(c) 2013-2020 Matt Hilton & Steven Boada

This is a higher level interface to some of the routines in PyWCSTools
(distributed with astLib).
PyWCSTools is a simple SWIG wrapping of WCSTools by Jessica Mink
(U{http://tdc-www.harvard.edu/software/wcstools/}). It is intended is to make
this interface complete enough such that direct use of PyWCSTools is
unnecessary.

"""

#-----------------------------------------------------------------------------

from astropy.io import fits as pyfits
from PyWCSTools import wcs
import astropy.wcs as apywcs
import numpy as np
import locale

# if True, -1 from pixel coords to be zero-indexed like numpy. If False, use
# FITS convention.
NUMPY_MODE = True
"""If True (default), pixel coordinates accepted/returned by routines such as
L{astWCS.WCS.pix2wcs}, L{astWCS.WCS.wcs2pix} have (0, 0) as the origin. Set
to False to make these routines accept/return pixel coords with (1, 1) as the
origin (i.e. to match the FITS convention, default behaviour prior to astLib
version 0.3.0)."""

# Check for the locale bug when decimal separator isn't '.' (atof used in
# libwcs)
lconv = locale.localeconv()
if lconv['decimal_point'] != '.':
    print("WARNING: decimal point separator is not '.' - astWCS coordinate conversions will not work.")
    print("Workaround: after importing any modules that set the locale (e.g. matplotlib) do the following:")
    print("   import locale")
    print("   locale.setlocale(locale.LC_NUMERIC, 'C')")

#-----------------------------------------------------------------------------
class WCS:
    """This class provides methods for accessing information from the World
    Coordinate System (WCS) contained in the header of a FITS image.
    Conversions between pixel and WCS coordinates can also be performed.

    To create a WCS object from a FITS file called "test.fits", simply:

    WCS=astWCS.WCS("test.fits")

    Likewise, to create a WCS object from the pyfits.header of "test.fits":

    img=pyfits.open("test.fits")
    header=img[0].header
    WCS=astWCS.WCS(header, mode = "pyfits")

    """

    def __init__(self, headerSource, extensionName = 0, mode = "image", zapKeywords = [],
                 useAstropyWCS = True, naxis = 2):
        """Creates a WCS object using either the information contained in the
        header of the specified .fits image, or from a pyfits.header object.
        Set mode = "pyfits" if the headerSource is a pyfits.header.

        For some images from some archives, particular header keywords such as 
        COMMENT or HISTORY may contain unprintable strings. If you encounter
        this, try setting zapKeywords = ['COMMENT', 'HISTORY'] (for example).
        
        @type headerSource: string or pyfits.header
        @param headerSource: filename of input .fits image, or a pyfits.header
            object
        @type extensionName: int or string
        @param extensionName: name or number of .fits extension in which image
            data is stored
        @type mode: string
        @param mode: set to "image" if headerSource is a .fits file name, or
            set to "astropy" if headerSource is an astropy.io.fits.Header object 
            (setting to this to "pyfits" is equivalent, for backwards 
            compatibility with existing code)
        @type zapKeywords: list
        @param: zapKeywords: keywords to remove from the header before making
            astWCS object.
        @type: useAstropyWCS: bool
        @param: useAstropyWCS: if True, use astropy.wcs to perform WCS 
            coordinate conversions in wcs2pix, pix2wcs (if False, use PyWCSTools)
            
        @note: The meta data provided by headerSource is stored in WCS.header
            as a pyfits.header object.

        """

        self.mode = mode
        self.headerSource = headerSource
        self.extensionName = extensionName
        self.naxis = naxis

        if self.mode == "image":
            img = pyfits.open(self.headerSource)
            # silentfix below won't deal with unprintable strings
            # so here we optionally remove problematic keywords
            for z in zapKeywords:
                if z in img[self.extensionName].header.keys():
                    for count in range(img[self.extensionName].header.count(z)):
                        img[self.extensionName].header.remove(z)
            img.verify('silentfix') # solves problems with non-standard headers
            self.header = img[self.extensionName].header
            img.close()
        elif self.mode == "astropy" or self.mode == "pyfits":
            for z in zapKeywords:
                if z in self.headerSource.keys():
                    for count in range(self.headerSource.count(z)):
                        self.headerSource.remove(z)
            self.header=headerSource
        
        # Scan for CUNIT values that upset astropy.wcs
        for i in (1, 2):
            if 'CUNIT%d' % (i) in self.header.keys():
                if self.header['CUNIT%d' % (i)] == '' or self.header['CUNIT%d' % (i)] == 'degree'\
                    or self.header['CUNIT%d' % (i)] == 'degrees' or self.header['CUNIT%d' % (i)] == 'DEG':
                    self.header['CUNIT%d' % (i)]='deg'
        
        # This enables a shim to allow code written for astLib to use astropy.wcs underneath
        self.useAstropyWCS=useAstropyWCS
        if NUMPY_MODE == True:
            self._apywcsOrigin=0
        else:
            self._apywcsOrigin=1

        self.updateFromHeader()


    def copy(self):
        """Copies the WCS object to a new object.

        @rtype: astWCS.WCS object
        @return: WCS object

        """

        # This only sets up a new WCS object, doesn't do a deep copy
        ret = WCS(self.headerSource, self.extensionName, self.mode, 
                  useAstropyWCS = self.useAstropyWCS)

        # This fixes copy bug
        ret.header = self.header.copy()
        ret.updateFromHeader()

        return ret


    def updateFromHeader(self):
        """Updates the WCS object using information from WCS.header. This
        routine should be called whenever changes are made to WCS keywords in
        WCS.header.

        """

        # Updated for pyfits 3.1+
        newHead=pyfits.Header()
        for i in self.header.items():
            if len(str(i[1])) < 70:
                if len(str(i[0])) <= 8:
                    newHead.append((i[0], i[1]))
                else:
                    newHead.append(('HIERARCH '+i[0], i[1]))
        
        # Workaround for ZPN bug when PV2_3 == 0 (as in, e.g., ESO WFI images)
        if "PV2_3" in list(newHead.keys()) and newHead['PV2_3'] == 0 and newHead['CTYPE1'] == 'RA---ZPN':
            newHead["PV2_3"]=1e-15
                
        cardstring = ""
        for card in newHead.cards:
            cardstring = cardstring+str(card)
        
        if self.useAstropyWCS == True:
            self.AWCS = apywcs.WCS(self.header, naxis = self.naxis) # For astropy.wcs shim
        self.WCSStructure = wcs.wcsinit(cardstring)


    def getCentreWCSCoords(self):
        """Returns the RA and dec coordinates (in decimal degrees) at the
        centre of the WCS.

        @rtype: list
        @return: coordinates in decimal degrees in format [RADeg, decDeg]

        """
        full = wcs.wcsfull(self.WCSStructure)
        RADeg = full[0]
        decDeg = full[1]

        return [RADeg, decDeg]


    def getFullSizeSkyDeg(self):
        """Returns the width, height of the image according to the WCS in
        decimal degrees on the sky (i.e., with the projection taken into
        account).

        @rtype: list
        @return: width and height of image in decimal degrees on the sky in
            format [width, height]

        """
        full = wcs.wcsfull(self.WCSStructure)
        width = full[2]
        height = full[3]

        return [width, height]


    def getHalfSizeDeg(self):
        """Returns the half-width, half-height of the image according to the
        WCS in RA and dec degrees.

        @rtype: list
        @return: half-width and half-height of image in R.A., dec. decimal
            degrees in format [half-width, half-height]

        """
        half = wcs.wcssize(self.WCSStructure)
        width = half[2]
        height = half[3]

        return [width, height]


    def getImageMinMaxWCSCoords(self):
        """Returns the minimum, maximum WCS coords defined by the size of the
        parent image (as defined by the NAXIS keywords in the image header).

        @rtype: list
        @return: [minimum R.A., maximum R.A., minimum Dec., maximum Dec.]

        """

        # Get size of parent image this WCS is taken from
        maxX = self.header['NAXIS1']
        maxY = self.header['NAXIS2']
        minX = 1.0
        minY = 1.0

        if NUMPY_MODE == True:
            maxX = maxX-1
            maxY = maxY-1
            minX = minX-1
            minY = minY-1

        bottomLeft = self.pix2wcs(minX, minY)
        topRight = self.pix2wcs(maxX, maxY)

        xCoords = [bottomLeft[0], topRight[0]]
        yCoords = [bottomLeft[1], topRight[1]]
        xCoords.sort()
        yCoords.sort()

        return [xCoords[0], xCoords[1], yCoords[0], yCoords[1]]


    def wcs2pix(self, RADeg, decDeg):
        """Returns the pixel coordinates corresponding to the input WCS
        coordinates (given in decimal degrees). RADeg, decDeg can be single
        floats, or lists or np arrays.

        @rtype: list
        @return: pixel coordinates in format [x, y]

        """
        if self.useAstropyWCS == False:
            if type(RADeg) == np.ndarray or type(RADeg) == list:
                if type(decDeg) == np.ndarray or type(decDeg) == list:
                    pixCoords = []
                    for ra, dec in zip(RADeg, decDeg):
                        pix = wcs.wcs2pix(self.WCSStructure, float(ra), float(dec))
                        # Below handles CEA wraparounds
                        if pix[0] < 1:
                            xTest = ((self.header['CRPIX1'])-(ra-360.0) /
                                self.getXPixelSizeDeg())
                            if xTest >= 1 and xTest < self.header['NAXIS1']:
                                pix[0] = xTest
                        if NUMPY_MODE == True:
                            pix[0] = pix[0]-1
                            pix[1] = pix[1]-1
                        pixCoords.append([pix[0], pix[1]])
            else:
                pixCoords = (wcs.wcs2pix(self.WCSStructure, float(RADeg),
                            float(decDeg)))
                # Below handles CEA wraparounds
                if pixCoords[0] < 1:
                    xTest = ((self.header['CRPIX1'])-(RADeg-360.0) /
                            self.getXPixelSizeDeg())
                    if xTest >= 1 and xTest < self.header['NAXIS1']:
                        pixCoords[0] = xTest
                if NUMPY_MODE == True:
                    pixCoords[0] = pixCoords[0]-1
                    pixCoords[1] = pixCoords[1]-1
                pixCoords = [pixCoords[0], pixCoords[1]]
        
        else:
            # astropy.wcs shim
            if self.naxis == 2:
                pixCoords = self.AWCS.all_world2pix(RADeg, decDeg, self._apywcsOrigin)
            elif self.naxis == 3:
                pixCoords = self.AWCS.all_world2pix(RADeg, decDeg, 0, self._apywcsOrigin)
            else:
                raise Exception("Not handling NAXIS > 3 with astropy.wcs shim")
            pixCoords = np.array(pixCoords)[:2].transpose().tolist()

        return pixCoords


    def pix2wcs(self, x, y):
        """Returns the WCS coordinates corresponding to the input pixel
        coordinates.

        @rtype: list
        @return: WCS coordinates in format [RADeg, decDeg]

        """
        if self.useAstropyWCS == False:
            if type(x) == np.ndarray or type(x) == list:
                if type(y) == np.ndarray or type(y) == list:
                    WCSCoords = []
                    for xc, yc in zip(x, y):
                        if NUMPY_MODE == True:
                            xc += 1
                            yc += 1
                        WCSCoords.append(wcs.pix2wcs(self.WCSStructure, float(xc),
                                    float(yc)))
            else:
                if NUMPY_MODE == True:
                    x += 1
                    y += 1
                WCSCoords = wcs.pix2wcs(self.WCSStructure, float(x), float(y))
        else:
            # astropy.wcs shim
            if self.naxis == 2:
                WCSCoords = self.AWCS.all_pix2world(x, y, self._apywcsOrigin)
            elif self.naxis == 3:
                WCSCoords = self.AWCS.all_pix2world(x, y, 0, self._apywcsOrigin)
            else:
                raise Exception("Not handling NAXIS > 3 with astropy.wcs shim")
            WCSCoords = np.array(WCSCoords)[:2].transpose().tolist()
            
        return WCSCoords


    def coordsAreInImage(self, RADeg, decDeg):
        """Returns True if the given RA, dec coordinate is within the image
        boundaries.

        @rtype: bool
        @return: True if coordinate within image, False if not.

        """
        
        pixCoords = self.wcs2pix(RADeg, decDeg)
        if pixCoords[0] >= 0 and pixCoords[0] < self.header['NAXIS1'] and \
            pixCoords[1] >= 0 and pixCoords[1] < self.header['NAXIS2']:
                return True
        else:
            return False


    def getRotationDeg(self):
        """Returns the rotation angle in degrees around the axis, North through
        East.

        @rtype: float
        @return: rotation angle in degrees

        """
        return self.WCSStructure.rot


    def isFlipped(self):
        """Returns 1 if image is reflected around axis, otherwise returns 0.

        @rtype: int
        @return: 1 if image is flipped, 0 otherwise

        """
        return self.WCSStructure.imflip


    def getPixelSizeDeg(self):
        """Returns the pixel scale of the WCS. This is the average of the x, y
        pixel scales.

        @rtype: float
        @return: pixel size in decimal degrees

        """

        avSize = (abs(self.WCSStructure.xinc)+abs(self.WCSStructure.yinc))/2.0

        return avSize


    def getXPixelSizeDeg(self):
        """Returns the pixel scale along the x-axis of the WCS in degrees.

        @rtype: float
        @return: pixel size in decimal degrees

        """

        avSize = abs(self.WCSStructure.xinc)

        return avSize


    def getYPixelSizeDeg(self):
        """Returns the pixel scale along the y-axis of the WCS in degrees.

        @rtype: float
        @return: pixel size in decimal degrees

        """

        avSize = abs(self.WCSStructure.yinc)

        return avSize


    def getEquinox(self):
        """Returns the equinox of the WCS.

        @rtype: float
        @return: equinox of the WCS

        """
        return self.WCSStructure.equinox


    def getEpoch(self):
        """Returns the epoch of the WCS.

        @rtype: float
        @return: epoch of the WCS

        """
        return self.WCSStructure.epoch


#-----------------------------------------------------------------------------
# Functions for comparing WCS objects
def findWCSOverlap(wcs1, wcs2):
    """Finds the minimum, maximum WCS coords that overlap between wcs1 and
    wcs2. Returns these coordinates, plus the corresponding pixel coordinates
    for each wcs. Useful for clipping overlapping region between two images.

    @rtype: dictionary
    @return: dictionary with keys 'overlapWCS' (min, max RA, dec of overlap
        between wcs1, wcs2) 'wcs1Pix', 'wcs2Pix' (pixel coords in each input
        WCS that correspond to 'overlapWCS' coords)

    """

    mm1 = wcs1.getImageMinMaxWCSCoords()
    mm2 = wcs2.getImageMinMaxWCSCoords()

    overlapWCSCoords = [0.0, 0.0, 0.0, 0.0]

    # Note order swapping below is essential
    # Min RA
    if mm1[0] - mm2[0] <= 0.0:
        overlapWCSCoords[0] = mm2[0]
    else:
        overlapWCSCoords[0] = mm1[0]

    # Max RA
    if mm1[1] - mm2[1] <= 0.0:
        overlapWCSCoords[1] = mm1[1]
    else:
        overlapWCSCoords[1] = mm2[1]

    # Min dec.
    if mm1[2] - mm2[2] <= 0.0:
        overlapWCSCoords[2] = mm2[2]
    else:
        overlapWCSCoords[2] = mm1[2]

    # Max dec.
    if mm1[3] - mm2[3] <= 0.0:
        overlapWCSCoords[3] = mm1[3]
    else:
        overlapWCSCoords[3] = mm2[3]

    # Get corresponding pixel coords
    p1Low = wcs1.wcs2pix(overlapWCSCoords[0], overlapWCSCoords[2])
    p1High = wcs1.wcs2pix(overlapWCSCoords[1], overlapWCSCoords[3])
    p1 = [p1Low[0], p1High[0], p1Low[1], p1High[1]]

    p2Low = wcs2.wcs2pix(overlapWCSCoords[0], overlapWCSCoords[2])
    p2High = wcs2.wcs2pix(overlapWCSCoords[1], overlapWCSCoords[3])
    p2 = [p2Low[0], p2High[0], p2Low[1], p2High[1]]

    return {'overlapWCS': overlapWCSCoords, 'wcs1Pix': p1, 'wcs2Pix': p2}

#-----------------------------------------------------------------------------
